from datetime import datetime, timezone
from time import time
from unittest import mock

from google.cloud.pubsub_v1.subscriber.message import Message as PubSubMessage
from google.pubsub_v1.types import ReceivedMessage


def build_gcp_queue_message(message):
    queue_message = mock.create_autospec(PubSubMessage, spec_set=True)
    payload, attrs = message.serialize()
    if isinstance(payload, str):
        payload = payload.encode('utf8')
        attrs['hedwig_encoding'] = 'utf8'
    queue_message.data, queue_message.attributes = payload, attrs
    queue_message.publish_time = datetime.now(timezone.utc)
    queue_message.delivery_attempt = 1
    return queue_message


def build_gcp_received_message(message):
    queue_message = mock.create_autospec(ReceivedMessage)
    queue_message.ack_id = "dummy_ack_id"
    payload, attrs = message.serialize()
    if isinstance(payload, str):
        payload = payload.encode('utf8')
        attrs['hedwig_encoding'] = 'utf8'
    queue_message.message = mock.create_autospec(PubSubMessage)
    queue_message.message.message_id = str(time())
    queue_message.message.data, queue_message.message.attributes = payload, attrs
    queue_message.message.publish_time = datetime.now(timezone.utc)
    queue_message.message.delivery_attempt = 1
    return queue_message



import abc
import logging
import threading
import uuid
from concurrent.futures import Future
from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import Optional, Union, Generator, List, Any, Dict, Tuple, Iterator

from hedwig.conf import settings
from hedwig.exceptions import ValidationError, IgnoreException, LoggingException, RetryException
from hedwig.models import Message
from hedwig.utils import log


class HedwigPublisherBaseBackend:
    @classmethod
    def topic(cls, message: Message) -> Union[str, Tuple[str, str]]:
        """
        The topic name for routing the message. When publishing cross-project topics, returned value may be a tuple of
        topic name and project id for Google or account id for AWS.
        """
        version_pattern = f'{message.major_version}.*'
        return settings.HEDWIG_MESSAGE_ROUTING[(message.type, version_pattern)]

    def _dispatch_sync(self, message: Message) -> None:
        from hedwig.backends.utils import get_consumer_backend

        consumer_backend = get_consumer_backend()
        queue_message = self._mock_queue_message(message)
        settings.HEDWIG_PRE_PROCESS_HOOK(**consumer_backend.pre_process_hook_kwargs(queue_message))
        consumer_backend.process_message(queue_message)
        settings.HEDWIG_POST_PROCESS_HOOK(**consumer_backend.post_process_hook_kwargs(queue_message))

    @abc.abstractmethod
    def _mock_queue_message(self, message: Message):
        """
        Generate a mock queue message in proper format as expected by the transport backend. This is primarily used for
        testing.
        """

    @abc.abstractmethod
    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> Union[str, Future]:
        """
        Actually publish a message with the given payload and attributes. Some transport mechanisms restrict payload
        to string formats, and others may require bytes. It's up to the implementing class to encode the payload
        properly. The implementing class is responsible for decoding the payload back to the same format (bytes or
        string) as serialized by the validator.
        """

    @contextmanager
    def _maybe_instrument(self, message: Message, instrumentation_headers: Dict) -> Iterator:
        try:
            import hedwig.instrumentation

            with hedwig.instrumentation.on_publish(message, instrumentation_headers) as span:
                yield span
        except ImportError:
            yield None

    def publish(self, message: Message) -> Union[str, Future]:
        """
        Publish a message
        :return: Either a future if publisher is async, message id otherwise. If a future is returned, it'll result in
        message id once completed
        """
        if settings.HEDWIG_SYNC:
            self._dispatch_sync(message)
            return str(uuid.uuid4())

        default_headers = settings.HEDWIG_DEFAULT_HEADERS(message=message)
        if default_headers:
            new_headers = {**default_headers, **message.headers}
            message = message.with_headers(new_headers)

        instrumentation_headers: Dict[str, str] = {}
        with self._maybe_instrument(message, instrumentation_headers):
            new_headers = {**message.headers, **instrumentation_headers}
            message = message.with_headers(new_headers)

            payload, attributes = message.serialize()

            result = self._publish(message, payload, attributes)

            log_published_message(message, result)

        return result


class HedwigConsumerBaseBackend:
    def __init__(self) -> None:
        self._error_count = 0
        self._heartbeat_called_at = datetime(1970, 1, 1)
        self._last_message_received_at = datetime(1970, 1, 1)
        self._heartbeat_interval_timedelta = timedelta(seconds=settings.HEDWIG_HEARTBEAT_INTERVAL_S)
        inactivity_reset_s = settings.HEDWIG_HEARTBEAT_INACTIVITY_RESET_S
        self._heartbeat_inactivity_reset_timedelta = None
        if inactivity_reset_s:
            self._heartbeat_inactivity_reset_timedelta = timedelta(seconds=inactivity_reset_s)

    def heartbeat_hook_kwargs(self) -> dict:
        return {"error_count": self.error_count}

    @staticmethod
    def pre_process_hook_kwargs(queue_message) -> dict:
        return {}

    @staticmethod
    def post_process_hook_kwargs(queue_message) -> dict:
        return {}

    @contextmanager
    def _maybe_instrument(self, **kwargs) -> Iterator:
        try:
            import hedwig.instrumentation

            with hedwig.instrumentation.on_receive(**kwargs) as span:
                yield span
        except ImportError:
            yield None

    def _maybe_update_instrumentation(self, message: Message) -> None:
        try:
            import hedwig.instrumentation

            hedwig.instrumentation.on_message(message)
        except ImportError:
            pass

    def message_handler(self, message_payload: Union[str, bytes], attributes: dict, provider_metadata) -> None:
        message = self._build_message(message_payload, attributes, provider_metadata)
        _log_received_message(message)

        self._maybe_update_instrumentation(message)

        message.exec_callback()

    def fetch_and_process_messages(
        self,
        num_messages: int = 10,
        visibility_timeout: Optional[int] = None,
        shutdown_event: Optional[threading.Event] = None,
    ) -> None:
        if not shutdown_event:
            shutdown_event = threading.Event()  # pragma: no cover
        while not shutdown_event.is_set():
            queue_messages = self.pull_messages(
                num_messages=num_messages, visibility_timeout=visibility_timeout, shutdown_event=shutdown_event
            )
            for queue_message in queue_messages:
                self._last_message_received_at = datetime.utcnow()
                with self._maybe_instrument(**self.pre_process_hook_kwargs(queue_message)):
                    try:
                        settings.HEDWIG_PRE_PROCESS_HOOK(**self.pre_process_hook_kwargs(queue_message))
                    except Exception:
                        log(
                            __name__,
                            logging.ERROR,
                            'Exception in pre process hook for message',
                            exc_info=True,
                            extra={'queue_message': queue_message},
                        )
                        self.nack_message(queue_message)
                        continue

                    try:
                        self.process_message(queue_message)
                        self._error_count = 0
                    except IgnoreException:
                        log(__name__, logging.INFO, 'Ignoring task', extra={'queue_message': queue_message})
                    except LoggingException as e:
                        # log with message and extra
                        log(__name__, logging.ERROR, str(e), extra=e.extra, exc_info=True)
                        self.nack_message(queue_message)
                        continue
                    except RetryException:
                        # Retry without logging exception
                        log(__name__, logging.INFO, 'Retrying due to exception')
                        self.nack_message(queue_message)
                        continue
                    except Exception:
                        log(__name__, logging.ERROR, 'Exception while processing message', exc_info=True)
                        self.nack_message(queue_message)
                        self._error_count += 1
                        continue
                    finally:
                        self._call_heartbeat_hook()

                    try:
                        settings.HEDWIG_POST_PROCESS_HOOK(**self.post_process_hook_kwargs(queue_message))
                    except Exception:
                        log(
                            __name__,
                            logging.ERROR,
                            'Exception in post process hook for message',
                            extra={'queue_message': queue_message},
                            exc_info=True,
                        )
                        self.nack_message(queue_message)
                        continue

                    try:
                        self.ack_message(queue_message)
                    except Exception:
                        log(
                            __name__,
                            logging.ERROR,
                            'Exception while deleting message',
                            extra={'queue_message': queue_message},
                            exc_info=True,
                        )

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata) -> None:
        """
        Extends visibility timeout of a message on a given priority queue for long running tasks.
        """
        raise NotImplementedError

    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: Optional[int] = None) -> None:
        """
        Re-queues everything in the Hedwig DLQ back into the Hedwig queue.

        :param num_messages: Maximum number of messages to fetch in one call. Defaults to 10.
        :param visibility_timeout: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
        """
        raise NotImplementedError

    def pull_messages(
        self,
        num_messages: int = 10,
        visibility_timeout: Optional[int] = None,
        shutdown_event: Optional[threading.Event] = None,
    ) -> Union[Generator, List]:
        """
        Pulls messages from the cloud for this app.
        :param shutdown_event:
        :param num_messages:
        :param visibility_timeout:
        :return: a tuple of list of messages and the queue they were pulled from
        """
        raise NotImplementedError

    def process_message(self, queue_message) -> None:
        raise NotImplementedError

    def process_messages(self, lambda_event) -> None:
        # for lambda backend
        raise NotImplementedError

    def ack_message(self, queue_message) -> None:
        raise NotImplementedError

    def nack_message(self, queue_message) -> None:
        raise NotImplementedError

    @staticmethod
    def _build_message(message_payload: Union[str, bytes], attributes: dict, provider_metadata: Any) -> Message:
        try:
            message = Message.deserialize(message_payload, attributes, provider_metadata)
            # side-effect: validates the callback
            _ = message.callback
            return message
        except ValidationError:
            _log_invalid_message(message_payload)
            raise

    def _perform_error_counter_inactivity_reset(self):
        if self._heartbeat_inactivity_reset_timedelta and self.error_count:
            now = datetime.utcnow()
            if self._last_message_received_at + self._heartbeat_inactivity_reset_timedelta < now:
                self._error_count = 0
                log(__name__, logging.INFO, 'Error counter was reset due to heartbeat inactivity settings')

    def _call_heartbeat_hook(self, force: bool = False):
        now = datetime.utcnow()
        if force or self._heartbeat_called_at + self._heartbeat_interval_timedelta < now:
            try:
                settings.HEDWIG_HEARTBEAT_HOOK(**self.heartbeat_hook_kwargs())
            except Exception:
                log(__name__, logging.ERROR, 'Exception in heartbeat hook', exc_info=True)
            finally:
                self._heartbeat_called_at = now

    @property
    def error_count(self) -> int:
        """
        Returns the number of consecutive errors occurred when trying to process messages from the queue.

        Resets to 0 when a message is successfully processed.

        :return: Number of consecutive errors
        """
        return self._error_count


def log_published_message(message: Message, result: Union[str, Future]) -> None:
    def _log(message_id: str):
        log(__name__, logging.DEBUG, 'Sent message', extra={'hedwig_message': message, 'message_id': message_id})

    if isinstance(result, Future):
        result.add_done_callback(lambda f: _log(f.result()))
    else:
        _log(result)


def _log_received_message(message: Message) -> None:
    log(__name__, logging.DEBUG, 'Received message', extra={'hedwig_message': message})


def _log_invalid_message(message_payload: Union[str, bytes]) -> None:
    log(__name__, logging.ERROR, 'Received invalid message', extra={'message_payload': message_payload})



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class TripCreatedV1(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    VEHICLE_ID_FIELD_NUMBER: builtins.int
    USER_ID_FIELD_NUMBER: builtins.int
    VIN_FIELD_NUMBER: builtins.int
    vehicle_id: typing.Text = ...
    user_id: typing.Text = ...
    vin: typing.Text = ...
    def __init__(self,
        *,
        vehicle_id : typing.Optional[typing.Text] = ...,
        user_id : typing.Optional[typing.Text] = ...,
        vin : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["user_id",b"user_id","vehicle_id",b"vehicle_id","vin",b"vin"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["user_id",b"user_id","vehicle_id",b"vehicle_id","vin",b"vin"]) -> None: ...
global___TripCreatedV1 = TripCreatedV1

class TripCreatedV2(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    VEHICLE_ID_FIELD_NUMBER: builtins.int
    VIN_FIELD_NUMBER: builtins.int
    vehicle_id: typing.Text = ...
    vin: typing.Text = ...
    def __init__(self,
        *,
        vehicle_id : typing.Optional[typing.Text] = ...,
        vin : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["vehicle_id",b"vehicle_id","vin",b"vin"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["vehicle_id",b"vehicle_id","vin",b"vin"]) -> None: ...
global___TripCreatedV2 = TripCreatedV2

class DeviceCreatedV1(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    DEVICE_ID_FIELD_NUMBER: builtins.int
    USER_ID_FIELD_NUMBER: builtins.int
    device_id: typing.Text = ...
    user_id: typing.Text = ...
    def __init__(self,
        *,
        device_id : typing.Optional[typing.Text] = ...,
        user_id : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["device_id",b"device_id","user_id",b"user_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["device_id",b"device_id","user_id",b"user_id"]) -> None: ...
global___DeviceCreatedV1 = DeviceCreatedV1

class VehicleCreatedV1(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    VEHICLE_ID_FIELD_NUMBER: builtins.int
    vehicle_id: typing.Text = ...
    def __init__(self,
        *,
        vehicle_id : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["vehicle_id",b"vehicle_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["vehicle_id",b"vehicle_id"]) -> None: ...
global___VehicleCreatedV1 = VehicleCreatedV1



import factory

from hedwig.testing.factories.base import BaseMessageFactory


class JSONSchemaDataFactory(factory.DictFactory):
    trip_id = 'T_1234567890123456'
    user_id = 'U_1234567890123456'
    device_id = 'abcdef00abcdef00abcdef00'
    vin = '00000000000000000'
    vehicle_id = 'C_1234567890123456'


class JSONSchemaMessageFactory(BaseMessageFactory):
    data = factory.SubFactory(JSONSchemaDataFactory)



from hedwig.conf import settings


def unconfigure() -> None:
    """
    If settings were configured, un-configure them - useful for testing only.
    """
    settings.clear_cache()
    settings._user_settings = None



import logging
from contextlib import contextmanager
from unittest import mock

import pytest

import hedwig.conf
from hedwig.backends.base import HedwigPublisherBaseBackend
from hedwig.backends.import_utils import import_module_attr
from hedwig.testing.config import unconfigure

from tests.models import MessageType

try:
    # may not be available
    from moto import mock_sqs, mock_sns
except ImportError:
    pass


def pytest_configure():
    logging.basicConfig()


@pytest.fixture
def settings():
    """
    Use this fixture to override settings. Changes are automatically reverted
    """
    hedwig.conf.settings._ensure_configured()
    original_module = hedwig.conf.settings._user_settings

    class Wrapped:
        # default to the original module, but allow tests to setattr which would override
        def __getattr__(self, name):
            return getattr(original_module, name)

    unconfigure()
    hedwig.conf.settings._user_settings = Wrapped()

    try:
        yield hedwig.conf.settings._user_settings
    finally:
        unconfigure()
        hedwig.conf.settings._user_settings = original_module


@pytest.fixture(name='message_factory', params=['jsonschema', 'protobuf'])
def _message_factory(request, settings):
    if request.param == 'jsonschema':
        settings.HEDWIG_DATA_VALIDATOR_CLASS = 'hedwig.validators.jsonschema.JSONSchemaValidator'

        try:
            import jsonschema  # noqa
            from hedwig.testing.factories.jsonschema import JSONSchemaMessageFactory  # noqa

            yield JSONSchemaMessageFactory
        except ImportError:
            pytest.skip("JSON Schema not importable")

    if request.param == 'protobuf':
        settings.HEDWIG_DATA_VALIDATOR_CLASS = 'hedwig.validators.protobuf.ProtobufValidator'

        try:
            from tests.protobuf_factory import ProtobufMessageFactory  # noqa

            def _encode_proto(msg):
                return msg.SerializeToString(deterministic=True)

            # make maps deterministically ordered
            with mock.patch("hedwig.validators.protobuf.ProtobufValidator._encode_proto", side_effect=_encode_proto):
                yield ProtobufMessageFactory
        except ImportError:
            pytest.skip("Protobuf factory not importable")


@pytest.fixture()
def message_data(message_factory):
    return message_factory.build(msg_type=MessageType.trip_created)


@pytest.fixture()
def message(message_factory):
    return message_factory(msg_type=MessageType.trip_created)


@pytest.fixture()
def message_with_trace(message_factory):
    return message_factory(
        msg_type=MessageType.trip_created,
        metadata__headers__traceparent="00-aa2ada259e917551e16da4a0ad33db24-662fd261d30ec74c-01",
    )


@contextmanager
def _mock_boto3():
    settings.AWS_REGION = 'us-west-1'
    with mock_sqs(), mock_sns(), mock.patch("hedwig.backends.aws.boto3", autospec=True) as boto3_mock:
        yield boto3_mock


@pytest.fixture
def mock_boto3():
    with _mock_boto3() as m:
        yield m


@pytest.fixture()
def sqs_consumer_backend(mock_boto3):
    # may not be available
    from hedwig.backends import aws

    yield aws.AWSSQSConsumerBackend()


@pytest.fixture
def mock_pubsub_v1():
    with mock.patch("hedwig.backends.gcp.pubsub_v1", autospec=True) as pubsub_v1_mock:
        yield pubsub_v1_mock


@pytest.fixture(params=['aws', 'google'])
def consumer_backend(request):
    if request.param == 'aws':
        try:
            from hedwig.backends.aws import AWSSQSConsumerBackend  # noqa

            with _mock_boto3():
                yield AWSSQSConsumerBackend()
        except ImportError:
            pytest.skip("AWS backend not importable")

    if request.param == 'google':
        try:
            from hedwig.backends.gcp import GooglePubSubConsumerBackend  # noqa

            with mock.patch("hedwig.backends.gcp.pubsub_v1"), mock.patch(
                "hedwig.backends.gcp.google_auth_default", return_value=(None, "DUMMY")
            ):
                yield GooglePubSubConsumerBackend()
        except ImportError:
            pytest.skip("Google backend not importable")


@pytest.fixture(
    params=["hedwig.backends.aws.AWSSNSConsumerBackend", "hedwig.backends.gcp.GooglePubSubPublisherBackend"]
)
def publisher_backend(request, mock_boto3):
    with mock.patch("hedwig.backends.gcp.pubsub_v1"):
        yield import_module_attr(request.param)


@pytest.fixture()
def mock_publisher_backend():
    with mock.patch.object(HedwigPublisherBaseBackend, '_publish'):
        yield HedwigPublisherBaseBackend()


@pytest.fixture(params=[True, False], ids=["message-attrs", "no-message-attrs"])
def use_transport_message_attrs(request, settings):
    settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES = request.param
    yield settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES



import typing
from concurrent.futures import Future
from typing import Optional

from hedwig.backends.base import HedwigPublisherBaseBackend
from hedwig.backends.utils import get_publisher_backend
from hedwig.models import Message


def publish(message: Message, backend: Optional[HedwigPublisherBaseBackend] = None) -> typing.Union[str, Future]:
    """
    Publishes a message on Hedwig topic
    :returns: for async publishers, returns a future that represents the publish api call, otherwise, returns
    the published message id
    """
    backend = backend or get_publisher_backend()
    return backend.publish(message)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedwig/protobuf/container.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedwig/protobuf/container.proto',
  package='hedwig',
  syntax='proto3',
  serialized_options=b'\n\034io.github.cloudchacho.hedwigZ)github.com/cloudchacho/hedwig-go/protobuf',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1fhedwig/protobuf/container.proto\x12\x06hedwig\x1a\x19google/protobuf/any.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb0\x01\n\nMetadataV1\x12\x11\n\tpublisher\x18\x01 \x01(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x07headers\x18\x03 \x03(\x0b\x32\x1f.hedwig.MetadataV1.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x89\x01\n\tPayloadV1\x12\x16\n\x0e\x66ormat_version\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12$\n\x08metadata\x18\x03 \x01(\x0b\x32\x12.hedwig.MetadataV1\x12\x0e\n\x06schema\x18\x04 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x05 \x01(\x0b\x32\x14.google.protobuf.AnyBI\n\x1cio.github.cloudchacho.hedwigZ)github.com/cloudchacho/hedwig-go/protobufb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_METADATAV1_HEADERSENTRY = _descriptor.Descriptor(
  name='HeadersEntry',
  full_name='hedwig.MetadataV1.HeadersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='hedwig.MetadataV1.HeadersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='hedwig.MetadataV1.HeadersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=234,
  serialized_end=280,
)

_METADATAV1 = _descriptor.Descriptor(
  name='MetadataV1',
  full_name='hedwig.MetadataV1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='publisher', full_name='hedwig.MetadataV1.publisher', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='hedwig.MetadataV1.timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='headers', full_name='hedwig.MetadataV1.headers', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_METADATAV1_HEADERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=280,
)


_PAYLOADV1 = _descriptor.Descriptor(
  name='PayloadV1',
  full_name='hedwig.PayloadV1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format_version', full_name='hedwig.PayloadV1.format_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='hedwig.PayloadV1.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='hedwig.PayloadV1.metadata', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='hedwig.PayloadV1.schema', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='hedwig.PayloadV1.data', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=283,
  serialized_end=420,
)

_METADATAV1_HEADERSENTRY.containing_type = _METADATAV1
_METADATAV1.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METADATAV1.fields_by_name['headers'].message_type = _METADATAV1_HEADERSENTRY
_PAYLOADV1.fields_by_name['metadata'].message_type = _METADATAV1
_PAYLOADV1.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['MetadataV1'] = _METADATAV1
DESCRIPTOR.message_types_by_name['PayloadV1'] = _PAYLOADV1
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MetadataV1 = _reflection.GeneratedProtocolMessageType('MetadataV1', (_message.Message,), {

  'HeadersEntry' : _reflection.GeneratedProtocolMessageType('HeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _METADATAV1_HEADERSENTRY,
    '__module__' : 'hedwig.protobuf.container_pb2'
    # @@protoc_insertion_point(class_scope:hedwig.MetadataV1.HeadersEntry)
    })
  ,
  'DESCRIPTOR' : _METADATAV1,
  '__module__' : 'hedwig.protobuf.container_pb2'
  # @@protoc_insertion_point(class_scope:hedwig.MetadataV1)
  })
_sym_db.RegisterMessage(MetadataV1)
_sym_db.RegisterMessage(MetadataV1.HeadersEntry)

PayloadV1 = _reflection.GeneratedProtocolMessageType('PayloadV1', (_message.Message,), {
  'DESCRIPTOR' : _PAYLOADV1,
  '__module__' : 'hedwig.protobuf.container_pb2'
  # @@protoc_insertion_point(class_scope:hedwig.PayloadV1)
  })
_sym_db.RegisterMessage(PayloadV1)


DESCRIPTOR._options = None
_METADATAV1_HEADERSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import base64
import random
import threading
from datetime import datetime, timezone
from unittest import mock

import pytest

from hedwig.conf import settings as hedwig_settings

try:
    from opentelemetry.trace import get_current_span, get_tracer
    from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

    from hedwig.backends.aws import AWSMetadata
    from hedwig.instrumentation.compat import get_hexadecimal_trace_id, id_generator_class, get_traceparent_string
except ImportError:
    pass

from tests.utils.mock import mock_return_once

aws = pytest.importorskip('hedwig.backends.aws')
instrumentation = pytest.importorskip('hedwig.instrumentation')


@pytest.fixture(name='sqs_consumer')
def _sqs_consumer(mock_boto3):
    return aws.AWSSQSConsumerBackend()


def test_fetch_and_process_message_follows_parent_trace(
    sqs_consumer,
    message_with_trace,
    use_transport_message_attrs,
):
    shutdown_event = threading.Event()
    num_messages = 3
    visibility_timeout = 4
    queue = mock.MagicMock()
    sqs_consumer.sqs_resource.get_queue_by_name = mock.MagicMock(return_value=queue)

    receipt = "receipt"
    sent_time = datetime.now(timezone.utc)
    first_receive_time = datetime.now(timezone.utc)
    receive_count = 1

    queue_message = mock.MagicMock()
    queue_message.receipt_handle = receipt
    payload, message_attributes = message_with_trace.serialize()
    if isinstance(payload, bytes):
        queue_message.body = base64.encodebytes(payload).decode()
        message_attributes['hedwig_encoding'] = 'base64'
    else:
        queue_message.body = payload
    queue_message.message_attributes = {
        k: {'DataType': 'String', 'StringValue': v} for k, v in message_attributes.items()
    }
    queue_message.attributes = {
        'ApproximateReceiveCount': receive_count,
        'SentTimestamp': int(sent_time.timestamp() * 1000),
        'ApproximateFirstReceiveTimestamp': int(first_receive_time.timestamp() * 1000),
    }

    def verify_span(*args, **kwargs):
        span = get_current_span()
        assert span.get_span_context().is_valid
        assert (
            get_hexadecimal_trace_id(span.get_span_context().trace_id)
            == message_with_trace.headers["traceparent"].split("-")[1]
        )
        return mock.DEFAULT

    mock_return_once(queue.receive_messages, [queue_message], [], shutdown_event)
    message_mock = mock.MagicMock()
    sqs_consumer._build_message = mock.MagicMock(return_value=message_mock)
    sqs_consumer.process_message = mock.MagicMock(wraps=sqs_consumer.process_message, side_effect=verify_span)
    sqs_consumer.message_handler = mock.MagicMock(wraps=sqs_consumer.message_handler)

    sqs_consumer.fetch_and_process_messages(num_messages, visibility_timeout, shutdown_event)

    sqs_consumer.sqs_resource.get_queue_by_name.assert_called_with(QueueName=sqs_consumer.queue_name)
    queue.receive_messages.assert_called_with(
        MaxNumberOfMessages=num_messages,
        MessageAttributeNames=['All'],
        AttributeNames=['All'],
        VisibilityTimeout=visibility_timeout,
        WaitTimeSeconds=sqs_consumer.WAIT_TIME_SECONDS,
    )
    sqs_consumer.process_message.assert_called_once_with(queue_message)
    sqs_consumer.message_handler.assert_called_once_with(
        payload,
        message_attributes,
        AWSMetadata(
            receipt,
            # truncates to millisecond precision
            first_receive_time.replace(microsecond=(first_receive_time.microsecond // 1000) * 1000),
            sent_time.replace(microsecond=(sent_time.microsecond // 1000) * 1000),
            receive_count,
        ),
    )
    message_mock.exec_callback.assert_called_once_with()
    queue_message.delete.assert_called_once_with()


def test_publish_sends_trace_id(mock_boto3, message, use_transport_message_attrs):
    sns_publisher = aws.AWSSNSPublisherBackend()
    queue = mock.MagicMock()
    sns_publisher.sns_client.publish.get_queue_by_name = mock.MagicMock(return_value=queue)

    span_id = random.getrandbits(64)

    tracer = get_tracer(__name__)
    with tracer.start_as_current_span(test_publish_sends_trace_id.__name__), mock.patch(
        f"{id_generator_class}.generate_span_id", return_value=span_id
    ):
        trace_id = get_current_span().get_span_context().trace_id
        sns_publisher.publish(message)

    mock_boto3.client.assert_called_once_with(
        'sns',
        region_name=hedwig_settings.AWS_REGION,
        aws_access_key_id=hedwig_settings.AWS_ACCESS_KEY,
        aws_secret_access_key=hedwig_settings.AWS_SECRET_KEY,
        aws_session_token=hedwig_settings.AWS_SESSION_TOKEN,
        endpoint_url=hedwig_settings.AWS_ENDPOINT_SQS,
        config=mock.ANY,
    )
    topic = sns_publisher._get_sns_topic(message)
    traceparent_string = get_traceparent_string(trace_id, span_id, 1)
    instrumentation_headers = {TraceContextTextMapPropagator._TRACEPARENT_HEADER_NAME: traceparent_string}
    message = message.with_headers({**message.headers, **instrumentation_headers})
    data, attrs = message.serialize()
    if not use_transport_message_attrs:
        attrs = message.headers
    if isinstance(data, bytes):
        data = base64.encodebytes(data).decode()
        attrs['hedwig_encoding'] = 'base64'
    sns_publisher.sns_client.publish.assert_called_once_with(
        TopicArn=topic,
        Message=data,
        MessageAttributes={k: {'DataType': 'String', 'StringValue': str(v)} for k, v in attrs.items()},
    )



import json
import math
from decimal import Decimal
import uuid

import pytest

pytest.importorskip('jsonschema')

from jsonschema import SchemaError  # noqa

from hedwig.exceptions import ValidationError  # noqa
from hedwig.validators.jsonschema import JSONSchemaValidator  # noqa
from hedwig.testing.factories.jsonschema import JSONSchemaMessageFactory  # noqa

from tests.models import MessageType  # noqa


class TestJSONDefault:
    def _validator(self):
        return JSONSchemaValidator()

    @pytest.mark.parametrize('value', [1469056316326, 1469056316326.123])
    def test__convert_to_json_decimal(self, value):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, data__decimal=Decimal(value))
        assert json.loads(self._validator().serialize(message)[0])['decimal'] == float(message.data['decimal'])

    @pytest.mark.parametrize('value', [math.nan, math.inf, -math.inf])
    def test__convert_to_json_disallow_nan(self, value, message_data):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, data__nan=value)
        with pytest.raises(ValueError):
            self._validator().serialize(message)

    def test__convert_to_json_non_serializable(self, message_data):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, data__obj=object())
        with pytest.raises(TypeError):
            self._validator().serialize(message)


class TestJSONSchemaValidator:
    def _validator(self):
        return JSONSchemaValidator()

    def test_constructor_checks_schema(self):
        schema = {}
        with pytest.raises(SchemaError):
            JSONSchemaValidator(schema)

    @pytest.mark.parametrize(
        'schema,schema_exc_error',
        [
            [{'schemas': []}, "Invalid schema file: expected key 'schemas' with non-empty value"],
            [
                {'schemas': {"device.created": []}},
                "Invalid definition for: 'device.created', value must contain a dict of valid versions",
            ],
            [
                {'schemas': {"device.created": {'fail-pattern': []}}},
                "Invalid version 'fail-pattern' for: 'device.created'",
            ],
            [{'schemas': {"device.created": {'1.*': []}}}, "Invalid schema for: 'device.created' '1.*'"],
            [
                {
                    'schemas': {
                        "device.created": {'1.*': {"x-version": "1.0"}},
                        "vehicle_created": {'1.*': {"x-version": "1.0"}},
                        "trip_created": {'1.*': {"x-version": "1.0"}},
                    }
                },
                "Schema not found for 'trip_created' v2.*",
            ],
            [
                {'schemas': {"trip_created": {'2.*': {}}}},
                "Invalid schema for: 'trip_created' '2.*': missing x-version",
            ],
            [
                {'schemas': {"trip_created": {'2.*': {"x-version": "foobar"}}}},
                "Invalid full version: 'foobar' for: 'trip_created' '2.*'",
            ],
            [
                {'schemas': {"trip_created": {'2.*': {"x-version": "1.1"}}}},
                "Invalid full version: '1.1' for: 'trip_created' '2.*'",
            ],
        ],
    )
    def test_check_schema(self, schema: dict, schema_exc_error):
        with pytest.raises(SchemaError) as exc_context:
            JSONSchemaValidator(schema)

        assert schema_exc_error in exc_context.value.args[0]

    def test_serialize(self, use_transport_message_attrs):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)
        if not use_transport_message_attrs:
            payload = {
                'format_version': '1.0',
                'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
                'id': message.id,
                'metadata': {
                    'timestamp': message.timestamp,
                    'publisher': message.publisher,
                    'headers': message.headers,
                },
                'data': message.data,
            }
            attributes = message.headers
            serialized = self._validator().serialize(message)
            assert (payload, attributes) == (json.loads(serialized[0]), serialized[1])
        else:
            payload = message.data
            attributes = {
                "hedwig_format_version": '1.0',
                "hedwig_schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0",
                "hedwig_id": message.id,
                "hedwig_publisher": message.publisher,
                "hedwig_message_timestamp": str(message.timestamp),
                **message.headers,
            }
            serialized = self._validator().serialize(message)
            assert (payload, attributes) == (json.loads(serialized[0]), serialized[1])

    def test_serialize_firehose(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect firehose serialization
        _ = use_transport_message_attrs
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)
        payload = {
            'format_version': '1.0',
            'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
            'id': message.id,
            'metadata': {'timestamp': message.timestamp, 'publisher': message.publisher, 'headers': message.headers},
            'data': message.data,
        }
        serialized = self._validator().serialize_firehose(message)
        assert payload == json.loads(serialized)

    def test_serialize_containerized(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect containerized serialization
        _ = use_transport_message_attrs
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)
        payload = {
            'format_version': '1.0',
            'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
            'id': message.id,
            'metadata': {'timestamp': message.timestamp, 'publisher': message.publisher, 'headers': message.headers},
            'data': message.data,
        }
        serialized = self._validator().serialize_containerized(message)
        assert payload == json.loads(serialized)

    def test_deserialize(self, use_transport_message_attrs):
        provider_metadata = object()
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)
        message = message.with_provider_metadata(provider_metadata)

        if not use_transport_message_attrs:
            message_payload = json.dumps(
                {
                    'format_version': '1.0',
                    'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
                    'id': message.id,
                    'metadata': {
                        'timestamp': message.timestamp,
                        'publisher': message.publisher,
                        'headers': message.headers,
                    },
                    'data': message.data,
                }
            )
            attributes = message.headers
        else:
            message_payload = json.dumps(message.data)
            attributes = {
                "hedwig_format_version": '1.0',
                "hedwig_schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0",
                "hedwig_id": message.id,
                "hedwig_publisher": message.publisher,
                "hedwig_message_timestamp": str(message.timestamp),
                **message.headers,
            }

        assert message == self._validator().deserialize(message_payload, attributes, provider_metadata)

    def test_deserialize_firehose(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect firehose deserialization
        _ = use_transport_message_attrs
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)

        message_payload = json.dumps(
            {
                'format_version': '1.0',
                'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
                'id': message.id,
                'metadata': {
                    'timestamp': message.timestamp,
                    'publisher': message.publisher,
                    'headers': message.headers,
                },
                'data': message.data,
            }
        )

        assert message == self._validator().deserialize_firehose(message_payload)

    def test_deserialize_containerized(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect containerized deserialization
        _ = use_transport_message_attrs
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=1)

        message_payload = json.dumps(
            {
                'format_version': '1.0',
                'schema': 'https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0',
                'id': message.id,
                'metadata': {
                    'timestamp': message.timestamp,
                    'publisher': message.publisher,
                    'headers': message.headers,
                },
                'data': message.data,
            }
        )

        assert message == self._validator().deserialize_containerized(message_payload)

    def test_deserialize_raises_error_invalid_schema(self):
        validator = self._validator()

        payload = '''{}'''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert 'Invalid message attribute: hedwig_format_version must be string, found: None' in str(e.value.args[0])

        payload = '{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "https://wrong.host/schema#/schemas/trip_created",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, attrs, None)
        assert e.value.args[0] == 'Invalid schema found: https://wrong.host/schema#/schemas/trip_created'

        payload = '{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "https://wrong.host/schema#/schemas/trip_created/1.0",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, attrs, None)
        assert e.value.args[0] == 'message schema must start with "https://github.com/cloudchacho/hedwig-python/schema"'

        payload = '{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "https://wrong.host/schema#/schemas/trip_created/9.0",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, attrs, None)
        assert e.value.args[0] == 'message schema must start with "https://github.com/cloudchacho/hedwig-python/schema"'

    def test_deserialize_raises_error_invalid_schema_container(self, settings):
        settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES = False

        validator = self._validator()

        payload = '''{}'''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert '"\'id\' is a required property"' in str(e.value.args[0])

        payload = '''{
            "format_version": "1.0",
            "schema": "https://wrong.host/schema#/schemas/trip_created",
            "id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "metadata": {
                "timestamp": 1,
                "publisher": "",
                "headers": {}
            },
            "data": {}
        }'''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert e.value.args[0] == 'Invalid schema found: https://wrong.host/schema#/schemas/trip_created'

        payload = '''{
            "format_version": "1.0",
            "schema": "https://wrong.host/schema#/schemas/trip_created/1.0",
            "id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "metadata": {
                "timestamp": 1,
                "publisher": "",
                "headers": {}
            },
            "data": {}
        }'''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert e.value.args[0] == 'message schema must start with "https://github.com/cloudchacho/hedwig-python/schema"'

        payload = '''{
            "format_version": "1.0",
            "schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/9.0",
            "id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "metadata": {
                "timestamp": 1,
                "publisher": "",
                "headers": {}
            },
            "data": {}
        }'''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert (
            e.value.args[0]
            == 'Definition not found in schema: https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/9.*'
        )

    def test_deserialize_fails_on_invalid_json(self, message_data):
        with pytest.raises(ValidationError):
            self._validator().deserialize("bad json", {}, None)

    def test_deserialize_raises_error_invalid_data(self):
        payload = '{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError):
            self._validator().deserialize(payload, attrs, None)

    def test_serialize_raises_error_invalid_schema(self):
        message = JSONSchemaMessageFactory(msg_type='foobar')
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert (
            e.value.args[0]
            == 'Definition not found in schema: https://github.com/cloudchacho/hedwig-python/schema#/schemas/foobar/1.*'
        )

    def test_serialize_raises_error_invalid_version(self):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, model_version=9)
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert (
            e.value.args[0]
            == 'Definition not found in schema: https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/9.*'
        )

    def test_serialize_raises_error_invalid_minor_version(self):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, addition_version=2)
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0] == 'Unknown minor version: 2, last known minor version: 1'

    def test_serialize_raises_error_invalid_data(self):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, data={})
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0][0].args[0] == "'vehicle_id' is a required property"

    def test_serialize_raises_error_invalid_headers(self):
        message = JSONSchemaMessageFactory(msg_type=MessageType.trip_created, metadata__headers__hedwig_foo="bar")
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0] == "Invalid header key: 'hedwig_foo' - can't begin with reserved namespace 'hedwig_'"

    def test_check_human_uuid(self):
        validator = self._validator()
        assert validator._check_human_uuid(str(uuid.uuid4()))
        assert validator._check_human_uuid('6cac5588-24cc-4b4f-bbf9-7dc0ce93f96e')
        assert validator._check_human_uuid('793befdf-56c2-416a-92ed-420e62b33eb5')

    def test_check_human_uuid_fails(self):
        validator = self._validator()
        assert not validator._check_human_uuid(uuid.uuid4().hex)
        assert not validator._check_human_uuid('6cac5588')
        assert not validator._check_human_uuid('yyyyyyyy-tttt-416a-92ed-420e62b33eb5')
        assert not validator._check_human_uuid(uuid.uuid4())


def test_custom_validator(settings):
    settings.HEDWIG_DATA_VALIDATOR_CLASS = 'tests.validator.CustomValidator'

    with pytest.raises(ValidationError):
        JSONSchemaMessageFactory(msg_type=MessageType.trip_created, addition_version=1, data__vin='o' * 17).serialize()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf_minor_versioned.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from hedwig.protobuf import options_pb2 as hedwig_dot_protobuf_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf_minor_versioned.proto',
  package='tests.minor_versioned',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1eprotobuf_minor_versioned.proto\x12\x15tests.minor_versioned\x1a\x1dhedwig/protobuf/options.proto\"O\n\rTripCreatedV1\x12\x12\n\nvehicle_id\x18\x01 \x02(\t\x12\x0f\n\x07user_id\x18\x02 \x02(\t\x12\x0b\n\x03vin\x18\x03 \x02(\t:\x0c\x9a\x82\x19\x02\x08\x01\x9a\x82\x19\x02\x10\x01\"8\n\rTripCreatedV2\x12\x12\n\nvehicle_id\x18\x01 \x02(\t\x12\x0b\n\x03vin\x18\x02 \x02(\t:\x06\x9a\x82\x19\x02\x08\x02\"=\n\x0f\x44\x65viceCreatedV1\x12\x11\n\tdevice_id\x18\x01 \x02(\t\x12\x0f\n\x07user_id\x18\x02 \x02(\t:\x06\x9a\x82\x19\x02\x08\x01\".\n\x10VehicleCreatedV1\x12\x12\n\nvehicle_id\x18\x01 \x02(\t:\x06\x9a\x82\x19\x02\x08\x01'
  ,
  dependencies=[hedwig_dot_protobuf_dot_options__pb2.DESCRIPTOR,])




_TRIPCREATEDV1 = _descriptor.Descriptor(
  name='TripCreatedV1',
  full_name='tests.minor_versioned.TripCreatedV1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vehicle_id', full_name='tests.minor_versioned.TripCreatedV1.vehicle_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='tests.minor_versioned.TripCreatedV1.user_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vin', full_name='tests.minor_versioned.TripCreatedV1.vin', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\001\232\202\031\002\020\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=88,
  serialized_end=167,
)


_TRIPCREATEDV2 = _descriptor.Descriptor(
  name='TripCreatedV2',
  full_name='tests.minor_versioned.TripCreatedV2',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vehicle_id', full_name='tests.minor_versioned.TripCreatedV2.vehicle_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vin', full_name='tests.minor_versioned.TripCreatedV2.vin', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\002',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=169,
  serialized_end=225,
)


_DEVICECREATEDV1 = _descriptor.Descriptor(
  name='DeviceCreatedV1',
  full_name='tests.minor_versioned.DeviceCreatedV1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_id', full_name='tests.minor_versioned.DeviceCreatedV1.device_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='tests.minor_versioned.DeviceCreatedV1.user_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=227,
  serialized_end=288,
)


_VEHICLECREATEDV1 = _descriptor.Descriptor(
  name='VehicleCreatedV1',
  full_name='tests.minor_versioned.VehicleCreatedV1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vehicle_id', full_name='tests.minor_versioned.VehicleCreatedV1.vehicle_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=290,
  serialized_end=336,
)

DESCRIPTOR.message_types_by_name['TripCreatedV1'] = _TRIPCREATEDV1
DESCRIPTOR.message_types_by_name['TripCreatedV2'] = _TRIPCREATEDV2
DESCRIPTOR.message_types_by_name['DeviceCreatedV1'] = _DEVICECREATEDV1
DESCRIPTOR.message_types_by_name['VehicleCreatedV1'] = _VEHICLECREATEDV1
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TripCreatedV1 = _reflection.GeneratedProtocolMessageType('TripCreatedV1', (_message.Message,), {
  'DESCRIPTOR' : _TRIPCREATEDV1,
  '__module__' : 'protobuf_minor_versioned_pb2'
  # @@protoc_insertion_point(class_scope:tests.minor_versioned.TripCreatedV1)
  })
_sym_db.RegisterMessage(TripCreatedV1)

TripCreatedV2 = _reflection.GeneratedProtocolMessageType('TripCreatedV2', (_message.Message,), {
  'DESCRIPTOR' : _TRIPCREATEDV2,
  '__module__' : 'protobuf_minor_versioned_pb2'
  # @@protoc_insertion_point(class_scope:tests.minor_versioned.TripCreatedV2)
  })
_sym_db.RegisterMessage(TripCreatedV2)

DeviceCreatedV1 = _reflection.GeneratedProtocolMessageType('DeviceCreatedV1', (_message.Message,), {
  'DESCRIPTOR' : _DEVICECREATEDV1,
  '__module__' : 'protobuf_minor_versioned_pb2'
  # @@protoc_insertion_point(class_scope:tests.minor_versioned.DeviceCreatedV1)
  })
_sym_db.RegisterMessage(DeviceCreatedV1)

VehicleCreatedV1 = _reflection.GeneratedProtocolMessageType('VehicleCreatedV1', (_message.Message,), {
  'DESCRIPTOR' : _VEHICLECREATEDV1,
  '__module__' : 'protobuf_minor_versioned_pb2'
  # @@protoc_insertion_point(class_scope:tests.minor_versioned.VehicleCreatedV1)
  })
_sym_db.RegisterMessage(VehicleCreatedV1)


_TRIPCREATEDV1._options = None
_TRIPCREATEDV2._options = None
_DEVICECREATEDV1._options = None
_VEHICLECREATEDV1._options = None
# @@protoc_insertion_point(module_scope)



import logging
from unittest import mock
from uuid import uuid4

import pytest

import hedwig.utils


@pytest.fixture(name="logger_mock", params=["structlog", "stdlog"])
def _logger_mock(request):
    if request.param == "stdlog":
        with mock.patch.object(hedwig.utils, "structlog", None), mock.patch("hedwig.utils.logging") as m:
            yield request.param, m
    elif request.param == "structlog":
        with mock.patch("hedwig.utils.structlog") as m:
            yield request.param, m


def test_log(logger_mock):
    mock_logger = logger_mock[1]
    logger = logger_mock[0]
    message = "foobar"
    hedwig.utils.log(__name__, logging.INFO, message)
    if logger == "stdlog":
        mock_logger.getLogger(__name__).log.assert_called_once_with(logging.INFO, message)
    elif logger == "structlog":
        mock_logger.getLogger(__name__).info.assert_called_once_with(message)


def test_log_exception(logger_mock):
    mock_logger = logger_mock[1]
    logger = logger_mock[0]
    message = "foobar"
    hedwig.utils.log(__name__, logging.ERROR, message, exc_info=True)
    if logger == "stdlog":
        mock_logger.getLogger(__name__).log.assert_called_once_with(logging.ERROR, message, exc_info=True)
    elif logger == "structlog":
        mock_logger.getLogger(__name__).error.assert_called_once_with(message, exc_info=True)


def test_log_extra(logger_mock):
    mock_logger = logger_mock[1]
    logger = logger_mock[0]
    message = "foobar"
    u = uuid4()
    hedwig.utils.log(__name__, logging.ERROR, message, extra={"uuid": str(u), "another": "field", "yet_another": 1})
    if logger == "stdlog":
        mock_logger.getLogger(__name__).log.assert_called_once_with(
            logging.ERROR, message, extra={"uuid": str(u), "another": "field", "yet_another": 1}
        )
    elif logger == "structlog":
        mock_logger.getLogger(__name__).error.assert_called_once_with(
            message, uuid=str(u), another="field", yet_another=1
        )



import base64
import dataclasses
import logging
import threading
from contextlib import contextmanager
from datetime import datetime, timezone
from time import time
from typing import cast, Optional, Generator, List, Union, Dict, Iterator
from unittest import mock

import boto3
import funcy
from botocore.config import Config
from retrying import retry

from hedwig.backends.base import HedwigConsumerBaseBackend, HedwigPublisherBaseBackend
from hedwig.backends.exceptions import PartialFailure
from hedwig.conf import settings
from hedwig.models import Message
from hedwig.utils import log


@dataclasses.dataclass(frozen=True)
class AWSMetadata:
    """
    AWS specific metadata for a Message
    """

    receipt: str
    """
    AWS receipt identifier
    """

    first_receive_time: datetime
    """
    The time the message was first received from the queue. The value
    is calculated as best effort and is approximate.
    """

    sent_time: datetime
    """
    Time this message was originally sent to AWS
    """

    receive_count: int
    """
    The receive count received from SQS.
    The first delivery of a given message will have this value as 1. The value
    is calculated as best effort and is approximate.
    """


class AWSSNSPublisherBackend(HedwigPublisherBaseBackend):
    def __init__(self):
        self._sns_client = None

    @property
    def sns_client(self):
        if self._sns_client is None:
            config = Config(connect_timeout=settings.AWS_CONNECT_TIMEOUT_S, read_timeout=settings.AWS_READ_TIMEOUT_S)
            self._sns_client = boto3.client(
                'sns',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SNS,
                config=config,
            )
        return self._sns_client

    @classmethod
    def _get_sns_topic(cls, message: Message) -> str:
        topic = cls.topic(message)
        if isinstance(topic, tuple):
            topic, account_id = topic
        else:
            account_id = settings.AWS_ACCOUNT_ID
        return f'arn:aws:sns:{settings.AWS_REGION}:{account_id}:hedwig-{topic}'

    @retry(stop_max_attempt_number=3, stop_max_delay=3000)
    def _publish_over_sns(self, topic: str, message_payload: str, attributes: Dict[str, str]) -> str:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sns.html#SNS.Topic.publish
        message_attributes = {str(k): {'DataType': 'String', 'StringValue': str(v)} for k, v in attributes.items()}
        response = self.sns_client.publish(
            TopicArn=topic, Message=message_payload, MessageAttributes=message_attributes
        )
        return response['MessageId']

    def _mock_queue_message(self, message: Message) -> mock.Mock:
        sqs_message = mock.Mock()
        payload, attributes = message.serialize()
        # SQS requires UTF-8 encoded string
        if isinstance(payload, bytes):
            payload = base64.encodebytes(payload).decode()
            attributes['hedwig_encoding'] = 'base64'
        sqs_message.body = payload
        sqs_message.message_attributes = {
            k: {'DataType': 'String', 'StringValue': str(v)} for k, v in attributes.items()
        }
        sqs_message.attributes = {
            'ApproximateReceiveCount': 1,
            'SentTimestamp': int(time() * 1000),
            'ApproximateFirstReceiveTimestamp': int(time() * 1000),
        }
        sqs_message.receipt_handle = 'test-receipt'
        return sqs_message

    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> str:
        topic = self._get_sns_topic(message)
        # SNS requires UTF-8 encoded string
        if isinstance(payload, bytes):
            payload = base64.encodebytes(payload).decode()
            attributes['hedwig_encoding'] = 'base64'
        return self._publish_over_sns(topic, payload, attributes)


class AWSSQSConsumerBackend(HedwigConsumerBaseBackend):
    WAIT_TIME_SECONDS = 20

    def __init__(self, dlq=False):
        super().__init__()
        self._sqs_resource = None
        self._sqs_client = None
        self.queue_name = f'HEDWIG-{settings.HEDWIG_QUEUE}{"-DLQ" if dlq else ""}'

    @property
    def sqs_resource(self):
        if self._sqs_resource is None:
            self._sqs_resource = boto3.resource(
                'sqs',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SQS,
            )
        return self._sqs_resource

    @property
    def sqs_client(self):
        if self._sqs_client is None:
            self._sqs_client = boto3.client(
                'sqs',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SQS,
            )
        return self._sqs_client

    def _get_queue(self):
        return self.sqs_resource.get_queue_by_name(QueueName=self.queue_name)

    def pull_messages(
        self,
        num_messages: int = 10,
        visibility_timeout: Optional[int] = None,
        shutdown_event: Optional[threading.Event] = None,
    ) -> Union[Generator, List]:
        """

        :param shutdown_event: Unused for this backend
        :param num_messages:
        :param visibility_timeout:
        :return:
        """
        params = {
            'MaxNumberOfMessages': num_messages,
            'WaitTimeSeconds': self.WAIT_TIME_SECONDS,
            'AttributeNames': ['All'],
            'MessageAttributeNames': ['All'],
        }
        if visibility_timeout is not None:
            params['VisibilityTimeout'] = visibility_timeout
        try:
            return self._get_queue().receive_messages(**params)
        finally:
            self._perform_error_counter_inactivity_reset()
            self._call_heartbeat_hook()

    def process_message(self, queue_message) -> None:
        attributes = {k: o['StringValue'] for k, o in (queue_message.message_attributes or {}).items()}
        # body is always UTF-8 string
        message_payload = queue_message.body
        if attributes.get("hedwig_encoding") == "base64":
            message_payload = base64.decodebytes(message_payload.encode())
        receipt = queue_message.receipt_handle
        self.message_handler(
            message_payload,
            attributes,
            AWSMetadata(
                receipt,
                datetime.fromtimestamp(
                    int(queue_message.attributes['ApproximateFirstReceiveTimestamp']) / 1000, tz=timezone.utc
                ),
                datetime.fromtimestamp(int(queue_message.attributes['SentTimestamp']) / 1000, tz=timezone.utc),
                int(queue_message.attributes['ApproximateReceiveCount']),
            ),
        )

    def ack_message(self, queue_message) -> None:
        queue_message.delete()

    def nack_message(self, queue_message) -> None:
        # let visibility timeout take care of it
        pass

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata: AWSMetadata) -> None:
        """
        Extends visibility timeout of a message on a given priority queue for long running tasks.
        """
        receipt = metadata.receipt
        queue_url = self.sqs_client.get_queue_url(QueueName=self.queue_name)['QueueUrl']
        self.sqs_client.change_message_visibility(
            QueueUrl=queue_url, ReceiptHandle=receipt, VisibilityTimeout=visibility_timeout_s
        )
        self._call_heartbeat_hook(force=True)

    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: Optional[int] = None) -> None:
        """
        Re-queues everything in the Hedwig DLQ back into the Hedwig queue.

        :param num_messages: Maximum number of messages to fetch in one SQS call. Defaults to 10.
        :param visibility_timeout: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
        """
        sqs_queue = self.sqs_resource.get_queue_by_name(QueueName=f'HEDWIG-{settings.HEDWIG_QUEUE}')
        dead_letter_queue = self._get_queue()

        log(__name__, logging.INFO, "Re-queueing messages from {} to {}".format(dead_letter_queue.url, sqs_queue.url))
        while True:
            queue_messages = self.pull_messages(num_messages=num_messages, visibility_timeout=visibility_timeout)
            queue_messages = cast(list, queue_messages)

            if not queue_messages:
                break

            log(__name__, logging.INFO, "got {} messages from dlq".format(len(queue_messages)))

            result = sqs_queue.send_messages(
                Entries=[
                    funcy.merge(
                        {'Id': queue_message.message_id, 'MessageBody': queue_message.body},
                        {'MessageAttributes': queue_message.message_attributes}
                        if queue_message.message_attributes
                        else {},
                    )
                    for queue_message in queue_messages
                ]
            )
            if result.get('Failed'):
                raise PartialFailure(result)

            dead_letter_queue.delete_messages(
                Entries=[
                    {'Id': message.message_id, 'ReceiptHandle': message.receipt_handle} for message in queue_messages
                ]
            )

            log(__name__, logging.INFO, "Re-queued {} messages".format(len(queue_messages)))

    @staticmethod
    def pre_process_hook_kwargs(queue_message) -> dict:
        return {'sqs_queue_message': queue_message}

    @staticmethod
    def post_process_hook_kwargs(queue_message) -> dict:
        return {'sqs_queue_message': queue_message}


class AWSSNSConsumerBackend(HedwigConsumerBaseBackend):
    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: Optional[int] = None) -> None:
        raise RuntimeError("invalid operation for backend")  # pragma: no cover

    def pull_messages(
        self,
        num_messages: int = 10,
        visibility_timeout: Optional[int] = None,
        shutdown_event: Optional[threading.Event] = None,
    ) -> Union[Generator, List]:
        raise RuntimeError("invalid operation for backend")  # pragma: no cover

    def ack_message(self, queue_message) -> None:
        raise RuntimeError("invalid operation for backend")  # pragma: no cover

    def nack_message(self, queue_message) -> None:
        raise RuntimeError("invalid operation for backend")  # pragma: no cover

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata) -> None:
        raise RuntimeError("invalid operation for backend")  # pragma: no cover

    @contextmanager
    def _maybe_instrument(self, **kwargs) -> Iterator:
        try:
            import hedwig.instrumentation

            with hedwig.instrumentation.on_receive(**kwargs) as span:
                yield span
        except ImportError:
            yield None

    def process_messages(self, lambda_event):
        for record in lambda_event['Records']:
            with self._maybe_instrument(sns_record=record):
                self.process_message(record)

    def process_message(self, queue_message) -> None:
        settings.HEDWIG_PRE_PROCESS_HOOK(sns_record=queue_message)
        message_payload = queue_message['Sns']['Message']
        attributes = {k: o['Value'] for k, o in queue_message['Sns']['MessageAttributes'].items()}
        if attributes.get("hedwig_encoding") == "base64":
            message_payload = base64.decodebytes(message_payload.encode()).decode()
        self.message_handler(message_payload, attributes, None)
        settings.HEDWIG_POST_PROCESS_HOOK(sns_record=queue_message)



import time
import typing
import uuid
from distutils.version import StrictVersion
from enum import Enum

import factory

from hedwig.conf import settings
from hedwig.models import Message, Metadata


class HeadersFactory(factory.DictFactory):
    request_id = factory.Faker('uuid4')


class MetadataFactory(factory.Factory):
    class Meta:
        model = Metadata

    timestamp = factory.LazyFunction(lambda: int(time.time() * 1000))
    publisher = factory.LazyFunction(lambda: settings.HEDWIG_PUBLISHER)
    headers = factory.SubFactory(HeadersFactory)

    @classmethod
    def _build(cls, model_class: typing.Type[Message], *args, **kwargs) -> typing.Dict:
        if args:
            raise ValueError("MessageFactory %r does not support Meta.inline_args.", cls)

        return kwargs


class BaseMessageFactory(factory.Factory):
    class Meta:
        model = Message
        abstract = True

    class Params:
        model_version = 1
        addition_version = 0
        msg_type = None  # required

    version = factory.LazyAttribute(lambda obj: StrictVersion(f'{obj.model_version}.{obj.addition_version}'))
    type = factory.LazyAttribute(lambda obj: obj.msg_type.value if isinstance(obj.msg_type, Enum) else obj.msg_type)
    id = factory.LazyFunction(lambda: str(uuid.uuid4()))
    metadata = factory.SubFactory(MetadataFactory)

    @classmethod
    def _build(cls, model_class: typing.Type[Message], *args, **kwargs) -> typing.Dict:
        if args:
            raise ValueError("MessageFactory %r does not support Meta.inline_args.", cls)

        return kwargs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf_bad.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from hedwig.protobuf import options_pb2 as hedwig_dot_protobuf_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf_bad.proto',
  package='tests.bad',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12protobuf_bad.proto\x12\ttests.bad\x1a\x1dhedwig/protobuf/options.proto\"\'\n\rDeviceCreated\x12\x0e\n\x06\x66oobar\x18\x01 \x02(\x05:\x06\x9a\x82\x19\x02\x08\x02\"<\n\rTripCreatedV3\x12\x11\n\tdevice_id\x18\x01 \x02(\t:\x18\x9a\x82\x19\x02\x08\x02\x9a\x82\x19\x0e\x1a\x0ctrip_created\"7\n\x0bTripCreated\x12\x0e\n\x06\x66oobar\x18\x01 \x02(\x05:\x18\x9a\x82\x19\x02\x08\x01\x9a\x82\x19\x0e\x1a\x0ctrip_created\"\x1f\n\rTripCreatedV4\x12\x0e\n\x06\x66oobar\x18\x01 \x02(\x05'
  ,
  dependencies=[hedwig_dot_protobuf_dot_options__pb2.DESCRIPTOR,])




_DEVICECREATED = _descriptor.Descriptor(
  name='DeviceCreated',
  full_name='tests.bad.DeviceCreated',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='foobar', full_name='tests.bad.DeviceCreated.foobar', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\002',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=64,
  serialized_end=103,
)


_TRIPCREATEDV3 = _descriptor.Descriptor(
  name='TripCreatedV3',
  full_name='tests.bad.TripCreatedV3',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_id', full_name='tests.bad.TripCreatedV3.device_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\002\232\202\031\016\032\014trip_created',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=165,
)


_TRIPCREATED = _descriptor.Descriptor(
  name='TripCreated',
  full_name='tests.bad.TripCreated',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='foobar', full_name='tests.bad.TripCreated.foobar', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\232\202\031\002\010\001\232\202\031\016\032\014trip_created',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=167,
  serialized_end=222,
)


_TRIPCREATEDV4 = _descriptor.Descriptor(
  name='TripCreatedV4',
  full_name='tests.bad.TripCreatedV4',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='foobar', full_name='tests.bad.TripCreatedV4.foobar', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=224,
  serialized_end=255,
)

DESCRIPTOR.message_types_by_name['DeviceCreated'] = _DEVICECREATED
DESCRIPTOR.message_types_by_name['TripCreatedV3'] = _TRIPCREATEDV3
DESCRIPTOR.message_types_by_name['TripCreated'] = _TRIPCREATED
DESCRIPTOR.message_types_by_name['TripCreatedV4'] = _TRIPCREATEDV4
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeviceCreated = _reflection.GeneratedProtocolMessageType('DeviceCreated', (_message.Message,), {
  'DESCRIPTOR' : _DEVICECREATED,
  '__module__' : 'protobuf_bad_pb2'
  # @@protoc_insertion_point(class_scope:tests.bad.DeviceCreated)
  })
_sym_db.RegisterMessage(DeviceCreated)

TripCreatedV3 = _reflection.GeneratedProtocolMessageType('TripCreatedV3', (_message.Message,), {
  'DESCRIPTOR' : _TRIPCREATEDV3,
  '__module__' : 'protobuf_bad_pb2'
  # @@protoc_insertion_point(class_scope:tests.bad.TripCreatedV3)
  })
_sym_db.RegisterMessage(TripCreatedV3)

TripCreated = _reflection.GeneratedProtocolMessageType('TripCreated', (_message.Message,), {
  'DESCRIPTOR' : _TRIPCREATED,
  '__module__' : 'protobuf_bad_pb2'
  # @@protoc_insertion_point(class_scope:tests.bad.TripCreated)
  })
_sym_db.RegisterMessage(TripCreated)

TripCreatedV4 = _reflection.GeneratedProtocolMessageType('TripCreatedV4', (_message.Message,), {
  'DESCRIPTOR' : _TRIPCREATEDV4,
  '__module__' : 'protobuf_bad_pb2'
  # @@protoc_insertion_point(class_scope:tests.bad.TripCreatedV4)
  })
_sym_db.RegisterMessage(TripCreatedV4)


_DEVICECREATED._options = None
_TRIPCREATEDV3._options = None
_TRIPCREATED._options = None
# @@protoc_insertion_point(module_scope)



import threading
from datetime import datetime, timedelta, timezone
from unittest import mock

import freezegun
import pytest

try:
    from google.cloud.pubsub_v1.types import FlowControl
except ImportError:
    pass

try:
    from hedwig.backends.gcp import GoogleMetadata
    from tests.utils.gcp import build_gcp_queue_message, build_gcp_received_message
except ImportError:
    pass
from hedwig.conf import settings
from hedwig.exceptions import ValidationError, CallbackNotFound

from tests.models import MessageType


gcp = pytest.importorskip('hedwig.backends.gcp')


@pytest.fixture(autouse=True)
def gcp_settings(settings):
    settings.GOOGLE_APPLICATION_CREDENTIALS = "DUMMY_GOOGLE_APPLICATION_CREDENTIALS"
    settings.HEDWIG_PUBLISHER_BACKEND = "hedwig.backends.gcp.GooglePubSubPublisherBackend"
    settings.HEDWIG_CONSUMER_BACKEND = "hedwig.backends.gcp.GooglePubSubConsumerBackend"
    settings.GOOGLE_CLOUD_PROJECT = "DUMMY_PROJECT_ID"
    settings.GOOGLE_PUBSUB_READ_TIMEOUT_S = 5
    settings.HEDWIG_QUEUE = settings.HEDWIG_QUEUE.lower()
    settings.HEDWIG_SUBSCRIPTIONS = ['topic1', 'topic2', ('topic3', 'other-project')]
    yield settings


class TestPubSubPublisher:
    def test__get_topic_path(self, mock_pubsub_v1, message_factory):
        gcp_publisher = gcp.GooglePubSubPublisherBackend()
        message = message_factory(msg_type=MessageType.vehicle_created)
        assert gcp_publisher._get_topic_path(message) == gcp_publisher.publisher.topic_path.return_value
        gcp_publisher.publisher.topic_path.assert_called_once_with(
            'project-id-or-account-id', 'hedwig-dev-vehicle-created-v1'
        )

    def test_publish_success(self, mock_pubsub_v1, message, gcp_settings, use_transport_message_attrs):
        gcp_publisher = gcp.GooglePubSubPublisherBackend()
        gcp_publisher.publisher.topic_path = mock.MagicMock(return_value="dummy_topic_path")

        message_id = gcp_publisher.publish(message)

        assert message_id == gcp_publisher.publisher.publish.return_value.result()

        payload, attributes = message.serialize()
        if not use_transport_message_attrs:
            attributes = message.headers
        if isinstance(payload, str):
            payload = payload.encode('utf8')
            attributes["hedwig_encoding"] = 'utf8'

        mock_pubsub_v1.PublisherClient.assert_called_once_with(batch_settings=())
        gcp_publisher.publisher.topic_path.assert_called_once_with(
            gcp_settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{gcp_publisher.topic(message)}'
        )
        gcp_publisher.publisher.publish.assert_called_once_with("dummy_topic_path", data=payload, **attributes)

    def test_sync_publish_success(self, mock_pubsub_v1, message, gcp_settings, use_transport_message_attrs):
        gcp_publisher = gcp.GooglePubSubAsyncPublisherBackend()
        gcp_publisher.publisher.topic_path = mock.MagicMock(return_value="dummy_topic_path")
        payload, attributes = message.serialize()

        future = gcp_publisher.publish(message)
        assert future == gcp_publisher.publisher.publish.return_value

        if not use_transport_message_attrs:
            attributes = message.headers
        if isinstance(payload, str):
            payload = payload.encode('utf8')
            attributes["hedwig_encoding"] = 'utf8'

        mock_pubsub_v1.PublisherClient.assert_called_once_with(batch_settings=())
        gcp_publisher.publisher.topic_path.assert_called_once_with(
            gcp_settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{gcp_publisher.topic(message)}'
        )
        gcp_publisher.publisher.publish.assert_called_once_with("dummy_topic_path", data=payload, **attributes)

    @freezegun.freeze_time()
    @mock.patch('tests.handlers._trip_created_handler', autospec=True)
    def test_sync_mode(self, callback_mock, mock_pubsub_v1, message, mock_publisher_backend, gcp_settings):
        gcp_settings.HEDWIG_SYNC = True

        receipt = 'test-receipt'
        publish_time = datetime.now(timezone.utc)
        delivery_attempt = 1

        message.publish()
        callback_mock.assert_called_once_with(
            message.with_provider_metadata(GoogleMetadata(receipt, 'test-subscription', publish_time, delivery_attempt))
        )

    def test_sync_mode_detects_invalid_callback(self, gcp_settings, mock_pubsub_v1, message_factory):
        gcp_settings.HEDWIG_SYNC = True

        message = message_factory(msg_type=MessageType.vehicle_created)
        with pytest.raises(ValidationError) as exc_info:
            message.publish()
        assert isinstance(exc_info.value.__context__, CallbackNotFound)


pre_process_hook = mock.MagicMock()
post_process_hook = mock.MagicMock()
heartbeat_hook = mock.MagicMock()


@pytest.fixture(name='subscription_paths')
def _subscription_paths(gcp_settings):
    return [mock.MagicMock() for _ in range(len(gcp_settings.HEDWIG_SUBSCRIPTIONS) + 1)]


@pytest.fixture(name='gcp_consumer')
def _gcp_consumer(mock_pubsub_v1, gcp_settings, subscription_paths):
    mock_pubsub_v1.SubscriberClient.subscription_path.side_effect = subscription_paths
    return gcp.GooglePubSubConsumerBackend()


@pytest.fixture(name='prepost_process_hooks')
def _prepost_process_hooks(gcp_settings):
    gcp_settings.HEDWIG_HEARTBEAT_HOOK = 'tests.test_backends.test_gcp.heartbeat_hook'
    gcp_settings.HEDWIG_PRE_PROCESS_HOOK = 'tests.test_backends.test_gcp.pre_process_hook'
    gcp_settings.HEDWIG_POST_PROCESS_HOOK = 'tests.test_backends.test_gcp.post_process_hook'
    yield
    pre_process_hook.reset_mock(side_effect=True)
    post_process_hook.reset_mock(side_effect=True)
    heartbeat_hook.reset_mock(side_effect=True)


class TestGCPConsumer:
    def test_pull_messages(self, mock_pubsub_v1, gcp_consumer, subscription_paths, prepost_process_hooks):
        shutdown_event = threading.Event()
        num_messages = 1
        visibility_timeout = 10
        messages = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(), mock.MagicMock()]
        futures = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(), mock.MagicMock()]

        def subscribe_side_effect(subscription_path, callback, flow_control, scheduler):
            # send message
            scheduler.schedule(None, message=messages[gcp_consumer.subscriber.subscribe.call_count - 1])

            if gcp_consumer.subscriber.subscribe.call_count == len(messages):
                shutdown_event.set()

            # return a "future"
            return futures[gcp_consumer.subscriber.subscribe.call_count - 1]

        gcp_consumer.subscriber.subscribe.side_effect = subscribe_side_effect

        messages_received = list(
            gcp_consumer.pull_messages(num_messages, visibility_timeout, shutdown_event=shutdown_event)
        )

        # unwrap messages
        assert [m.message for m in messages_received] == messages

        # should look up all subscription paths correctly
        mock_pubsub_v1.SubscriberClient.subscription_path.assert_has_calls(
            [
                mock.call(
                    settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{settings.HEDWIG_QUEUE}-{settings.HEDWIG_SUBSCRIPTIONS[0]}'
                ),
                mock.call(
                    settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{settings.HEDWIG_QUEUE}-{settings.HEDWIG_SUBSCRIPTIONS[1]}'
                ),
                mock.call(
                    settings.GOOGLE_CLOUD_PROJECT,
                    f'hedwig-{settings.HEDWIG_QUEUE}-{settings.HEDWIG_SUBSCRIPTIONS[2][1]}-'
                    f'{settings.HEDWIG_SUBSCRIPTIONS[2][0]}',
                ),
                mock.call(settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{settings.HEDWIG_QUEUE}'),
            ]
        )

        # futures must be canceled on shutdown
        for future in futures:
            future.cancel.assert_called_once_with()

        # fetch the right number of messages
        flow_control = FlowControl(max_messages=num_messages, max_duration_per_lease_extension=visibility_timeout)

        # verify subscriber call for each path
        gcp_consumer.subscriber.subscribe.assert_has_calls(
            [
                mock.call(subscription_paths[x], callback=None, flow_control=flow_control, scheduler=mock.ANY)
                for x in range(4)
            ]
        )
        heartbeat_hook.assert_called_once_with(error_count=0)

    @pytest.mark.parametrize(
        "inactivity_s,last_message_received_delta_s,expected_error_count",
        [(None, 0, 1), (10, 1, 1), (1, 2, 0)],
        ids=["reset-disabled", "reset-enabled-inactivity-period-not-reached", "reset-enabled"],
    )
    def test_pull_messages_error_count_inactivity_reset(
        self,
        mock_pubsub_v1,
        gcp_settings,
        subscription_paths,
        prepost_process_hooks,
        inactivity_s,
        last_message_received_delta_s,
        expected_error_count,
    ):
        class _Event(threading.Event):
            def __init__(self, max_calls):
                super().__init__()
                self.max_calls = max_calls
                self.counter = 0

            def is_set(self):
                if self.counter == self.max_calls:
                    self.set()
                self.counter += 1
                return super().is_set()

        shutdown_event = _Event(max_calls=1)
        num_messages = 1
        visibility_timeout = 10
        mock_pubsub_v1.SubscriberClient.subscription_path.side_effect = subscription_paths
        gcp_settings.HEDWIG_HEARTBEAT_INACTIVITY_RESET_S = inactivity_s
        gcp_consumer = gcp.GooglePubSubConsumerBackend()
        gcp_consumer._last_message_received_at = datetime.utcnow() - timedelta(seconds=last_message_received_delta_s)
        gcp_consumer._error_count = 1

        list(gcp_consumer.pull_messages(num_messages, visibility_timeout, shutdown_event=shutdown_event))

        assert gcp_consumer._error_count == expected_error_count
        heartbeat_hook.assert_called_once_with(error_count=expected_error_count)

    def test_success_extend_visibility_timeout(self, gcp_consumer, prepost_process_hooks):
        visibility_timeout_s = 10
        ack_id = "dummy_ack_id"
        subscription_path = "subscriptions/foobar"
        publish_time = datetime.now(timezone.utc)
        delivery_attempt = 1

        gcp_consumer.extend_visibility_timeout(
            visibility_timeout_s, GoogleMetadata(ack_id, subscription_path, publish_time, delivery_attempt)
        )

        gcp_consumer.subscriber.modify_ack_deadline.assert_called_once_with(
            subscription=subscription_path, ack_ids=[ack_id], ack_deadline_seconds=visibility_timeout_s
        )
        heartbeat_hook.assert_called_once_with(error_count=0)

    @pytest.mark.parametrize("visibility_timeout", [-1, 601])
    def test_failure_extend_visibility_timeout(self, visibility_timeout, gcp_consumer, prepost_process_hooks):
        subscription_path = "subscriptions/foobar"
        publish_time = datetime.now(timezone.utc)
        delivery_attempt = 1

        with pytest.raises(ValueError):
            gcp_consumer.extend_visibility_timeout(
                visibility_timeout, GoogleMetadata('dummy_ack_id', subscription_path, publish_time, delivery_attempt)
            )

        gcp_consumer.subscriber.subscription_path.assert_not_called()
        gcp_consumer.subscriber.modify_ack_deadline.assert_not_called()
        heartbeat_hook.assert_not_called()

    def test_success_requeue_dead_letter(self, mock_pubsub_v1, message, use_transport_message_attrs):
        gcp_consumer = gcp.GooglePubSubConsumerBackend(dlq=True)

        num_messages = 1
        visibility_timeout = 4
        subscription_path = gcp_consumer._subscription_paths[0]

        queue_message = build_gcp_received_message(message)
        response = mock.MagicMock()
        response.received_messages = [queue_message]
        response2 = mock.MagicMock()
        response2.received_messages = []
        gcp_consumer.subscriber.pull.side_effect = iter([response, response2])

        gcp_consumer.requeue_dead_letter(num_messages=num_messages, visibility_timeout=visibility_timeout)

        gcp_consumer.subscriber.modify_ack_deadline.assert_called_once_with(
            subscription=subscription_path, ack_ids=[queue_message.ack_id], ack_deadline_seconds=visibility_timeout
        )
        gcp_consumer.subscriber.pull.assert_has_calls(
            [
                mock.call(
                    subscription=subscription_path,
                    max_messages=num_messages,
                    retry=None,
                    timeout=settings.GOOGLE_PUBSUB_READ_TIMEOUT_S,
                ),
                mock.call(
                    subscription=subscription_path,
                    max_messages=num_messages,
                    retry=None,
                    timeout=settings.GOOGLE_PUBSUB_READ_TIMEOUT_S,
                ),
            ]
        )
        gcp_consumer._publisher.publish.assert_called_once_with(
            mock_pubsub_v1.PublisherClient.topic_path.return_value,
            data=queue_message.message.data,
            **queue_message.message.attributes,
        )
        gcp_consumer.subscriber.acknowledge.assert_called_once_with(
            subscription=subscription_path, ack_ids=[queue_message.ack_id]
        )

    def test_fetch_and_process_messages_success(
        self,
        gcp_consumer,
        message,
        subscription_paths,
        prepost_process_hooks,
        use_transport_message_attrs,
    ):
        shutdown_event = threading.Event()
        num_messages = 3
        visibility_timeout = 4

        queue_message = build_gcp_queue_message(message)

        def subscribe_side_effect(subscription_path, callback, flow_control, scheduler):
            if gcp_consumer.subscriber.subscribe.call_count == 1:
                # send message
                scheduler.schedule(None, message=queue_message)

            shutdown_event.set()

            # return a "future"
            return mock.MagicMock()

        gcp_consumer.subscriber.subscribe.side_effect = subscribe_side_effect
        gcp_consumer.process_message = mock.MagicMock(wraps=gcp_consumer.process_message)
        gcp_consumer.message_handler = mock.MagicMock(wraps=gcp_consumer.message_handler)

        gcp_consumer.fetch_and_process_messages(
            num_messages=num_messages, visibility_timeout=visibility_timeout, shutdown_event=shutdown_event
        )

        # fetch the right number of messages
        flow_control = FlowControl(max_messages=num_messages, max_duration_per_lease_extension=visibility_timeout)

        gcp_consumer.subscriber.subscribe.assert_called_with(
            subscription_paths[-1], callback=None, flow_control=flow_control, scheduler=mock.ANY
        )
        gcp_consumer.process_message.assert_called_once_with(mock.ANY)
        assert gcp_consumer.process_message.call_args[0][0].message == queue_message
        if queue_message.attributes.get('hedwig_encoding') == 'utf8':
            payload = queue_message.data.decode('utf8')
        else:
            payload = queue_message.data
        gcp_consumer.message_handler.assert_called_once_with(
            payload,
            queue_message.attributes,
            GoogleMetadata(
                queue_message.ack_id,
                subscription_paths[0],
                queue_message.publish_time,
                queue_message.delivery_attempt,
            ),
        )
        queue_message.ack.assert_called_once_with()
        pre_process_hook.assert_called_once_with(google_pubsub_message=queue_message)
        post_process_hook.assert_called_once_with(google_pubsub_message=queue_message)
        heartbeat_hook.assert_called_once_with(error_count=0)



import random
import threading
from unittest import mock

import pytest

try:
    from opentelemetry.trace import get_current_span, get_tracer
    from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

    from hedwig.instrumentation.compat import get_hexadecimal_trace_id, id_generator_class, get_traceparent_string

    from tests.utils.gcp import build_gcp_queue_message
except ImportError:
    pass

gcp = pytest.importorskip('hedwig.backends.gcp')
instrumentation = pytest.importorskip('hedwig.instrumentation')


@pytest.fixture(name='subscription_paths')
def _subscription_paths(gcp_settings):
    return [mock.MagicMock() for _ in range(len(gcp_settings.HEDWIG_SUBSCRIPTIONS) + 1)]


@pytest.fixture(name='gcp_consumer')
def _gcp_consumer(mock_pubsub_v1, gcp_settings, subscription_paths):
    mock_pubsub_v1.SubscriberClient.subscription_path.side_effect = subscription_paths
    return gcp.GooglePubSubConsumerBackend()


@pytest.fixture(autouse=True)
def gcp_settings(settings):
    settings.GOOGLE_APPLICATION_CREDENTIALS = "DUMMY_GOOGLE_APPLICATION_CREDENTIALS"
    settings.HEDWIG_PUBLISHER_BACKEND = "hedwig.backends.gcp.GooglePubSubPublisherBackend"
    settings.HEDWIG_CONSUMER_BACKEND = "hedwig.backends.gcp.GooglePubSubConsumerBackend"
    settings.GOOGLE_CLOUD_PROJECT = "DUMMY_PROJECT_ID"
    settings.GOOGLE_PUBSUB_READ_TIMEOUT_S = 5
    settings.HEDWIG_QUEUE = settings.HEDWIG_QUEUE.lower()
    settings.HEDWIG_SUBSCRIPTIONS = ['topic1', 'topic2', ('topic3', 'other-project')]
    yield settings


def test_fetch_and_process_message_follows_parent_trace(
    gcp_consumer,
    message_with_trace,
    use_transport_message_attrs,
):
    shutdown_event = threading.Event()
    num_messages = 3
    visibility_timeout = 4

    queue_message = build_gcp_queue_message(message_with_trace)

    def subscribe_side_effect(subscription_path, callback, flow_control, scheduler):
        if gcp_consumer.subscriber.subscribe.call_count == 1:
            # send message
            scheduler.schedule(None, message=queue_message)

        shutdown_event.set()

        # return a "future"
        return mock.MagicMock()

    def verify_span(*args, **kwargs):
        curr_span = get_current_span()
        assert curr_span.get_span_context().is_valid
        assert (
            get_hexadecimal_trace_id(curr_span.get_span_context().trace_id)
            == message_with_trace.headers["traceparent"].split("-")[1]
        )
        return mock.DEFAULT

    gcp_consumer.subscriber.subscribe.side_effect = subscribe_side_effect
    gcp_consumer.process_message = mock.MagicMock(wraps=gcp_consumer.process_message, side_effect=verify_span)
    gcp_consumer.message_handler = mock.MagicMock(wraps=gcp_consumer.message_handler)

    gcp_consumer.fetch_and_process_messages(
        num_messages=num_messages, visibility_timeout=visibility_timeout, shutdown_event=shutdown_event
    )

    gcp_consumer.message_handler.assert_called_once()
    queue_message.ack.assert_called_once_with()


def test_publish_sends_trace_id(mock_pubsub_v1, message, gcp_settings, use_transport_message_attrs):
    gcp_publisher = gcp.GooglePubSubPublisherBackend()
    gcp_publisher.publisher.topic_path = mock.MagicMock(return_value="dummy_topic_path")

    span_id = random.getrandbits(64)

    tracer = get_tracer(__name__)
    with tracer.start_as_current_span(test_publish_sends_trace_id.__name__), mock.patch(
        f"{id_generator_class}.generate_span_id", return_value=span_id
    ):
        trace_id = get_current_span().get_span_context().trace_id
        message_id = gcp_publisher.publish(message)

    assert message_id == gcp_publisher.publisher.publish.return_value.result()

    traceparent_string = get_traceparent_string(trace_id, span_id, 1)
    instrumentation_headers = {TraceContextTextMapPropagator._TRACEPARENT_HEADER_NAME: traceparent_string}
    message = message.with_headers({**message.headers, **instrumentation_headers})
    payload, attributes = message.serialize()
    if not use_transport_message_attrs:
        attributes = message.headers
    if isinstance(payload, str):
        payload = payload.encode('utf8')
        attributes["hedwig_encoding"] = 'utf8'

    mock_pubsub_v1.PublisherClient.assert_called_once_with(batch_settings=())
    gcp_publisher.publisher.topic_path.assert_called_once_with(
        gcp_settings.GOOGLE_CLOUD_PROJECT, f'hedwig-{gcp_publisher.topic(message)}'
    )
    gcp_publisher.publisher.publish.assert_called_once_with("dummy_topic_path", data=payload, **attributes)



from enum import Enum

import factory

from hedwig.testing.factories.base import BaseMessageFactory
from hedwig.validators.protobuf import ProtobufValidator


def data_creator(o):
    msg_type = o.msg_type
    if isinstance(msg_type, Enum):
        msg_type = msg_type.value
    msg_class = ProtobufValidator().proto_messages[(msg_type, o.model_version)]
    msg = msg_class()
    for attr in ('trip_id', 'user_id', 'device_id', 'vin', 'vehicle_id'):
        try:
            setattr(msg, attr, getattr(o, attr))
        except AttributeError:
            # field not present in msg
            pass
    return msg


class ProtobufMessageFactory(BaseMessageFactory):
    class Params:
        trip_id = 'T_1234567890123456'
        user_id = 'U_1234567890123456'
        device_id = 'abcdef00abcdef00abcdef00'
        vin = '00000000000000000'
        vehicle_id = 'C_1234567890123456'
        protobuf_schema_module = None  # required

    data = factory.LazyAttribute(data_creator)



from contextlib import contextmanager
from typing import Dict, Optional, Iterator

from opentelemetry import trace
from opentelemetry.trace import Span

from hedwig.instrumentation.compat import Getter, extract, inject
from hedwig.models import Message

getter = Getter()


@contextmanager
def on_receive(sns_record=None, sqs_queue_message=None, google_pubsub_message=None) -> Iterator[Span]:
    """
    Hook for instrumenting consumer after message is dequeued. If applicable, starts a new span.
    :param sns_record:
    :param sqs_queue_message:
    :param google_pubsub_message:
    :return:
    """
    attributes: Optional[Dict]
    if sqs_queue_message is not None:
        attributes = {k: v["StringValue"] for k, v in sqs_queue_message.message_attributes.items()}
    elif sns_record is not None:
        attributes = sns_record["attributes"]
    elif google_pubsub_message is not None:
        attributes = google_pubsub_message.attributes
    else:
        attributes = None
    tracectx = extract(getter, attributes)  # type: ignore

    tracer = trace.get_tracer(__name__)
    with tracer.start_as_current_span("message_received", context=tracectx, kind=trace.SpanKind.CONSUMER) as span:
        yield span


def on_message(message: Message) -> None:
    """
    Hook for instrumenting consumer after message is deserialized and validated. If applicable, updates the current span
    with the right name.
    :param message:
    :return:
    """
    span = trace.get_current_span()
    span.update_name(message.type)


@contextmanager
def on_publish(message: Message, headers: Dict) -> Iterator[Span]:
    """
    Hook for instrumenting publish. If applicable, injects tracing headers into headers dictionary.
    :param message:
    :param headers:
    :return:
    """
    tracer = trace.get_tracer(__name__)
    with tracer.start_as_current_span(f"publish/{message.type}", kind=trace.SpanKind.PRODUCER) as span:
        inject(dict.__setitem__, headers)
        yield span



import json
import logging
import threading
from unittest import mock

import pytest

from hedwig.backends.base import HedwigConsumerBaseBackend, HedwigPublisherBaseBackend
from hedwig.backends.utils import get_consumer_backend, get_publisher_backend
from hedwig.conf import settings
from hedwig.models import ValidationError
from hedwig.exceptions import LoggingException, RetryException, IgnoreException
from tests.utils.mock import mock_return_once


class MockBackend(HedwigConsumerBaseBackend, HedwigPublisherBaseBackend):
    pass


class TestBackends:
    def test_success_get_consumer_backend(self, settings):
        settings.HEDWIG_CONSUMER_BACKEND = "tests.test_backends.test_base.MockBackend"

        consumer_backend = get_consumer_backend()

        assert isinstance(consumer_backend, MockBackend)

    def test_success_get_publisher_backend(self, settings):
        settings.HEDWIG_PUBLISHER_BACKEND = "tests.test_backends.test_base.MockBackend"

        publisher_backend = get_publisher_backend()

        assert isinstance(publisher_backend, MockBackend)

    @pytest.mark.parametrize("get_backend_fn", [get_publisher_backend, get_consumer_backend])
    def test_failure(self, get_backend_fn, settings):
        settings.HEDWIG_PUBLISHER_BACKEND = settings.HEDWIG_CONSUMER_BACKEND = "hedwig.backends.invalid"

        with pytest.raises(ImportError):
            get_backend_fn()


@mock.patch('hedwig.backends.base.Message.exec_callback', autospec=True)
class TestMessageHandler:
    def test_success(self, mock_exec_callback, message, consumer_backend, use_transport_message_attrs):
        provider_metadata = mock.Mock()
        consumer_backend.message_handler(*message.serialize(), provider_metadata)
        mock_exec_callback.assert_called_once_with(message.with_provider_metadata(provider_metadata))

    def test_fails_on_validation_error(self, mock_exec_callback, message, consumer_backend):
        if settings.HEDWIG_DATA_VALIDATOR_CLASS.__name__ == 'ProtobufValidator':
            m = mock.patch(
                'hedwig.validators.protobuf.ProtobufValidator.deserialize',
                new=mock.MagicMock(autospec=True, side_effect=ValidationError('Invalid message body')),
            )
        else:
            m = mock.patch(
                'hedwig.validators.jsonschema.JSONSchemaValidator.deserialize',
                new=mock.MagicMock(autospec=True, side_effect=ValidationError('Invalid message body')),
            )

        with m, pytest.raises(ValidationError):
            consumer_backend.message_handler(*message.serialize(), None)
        mock_exec_callback.assert_not_called()

    def test_fails_on_task_failure(self, mock_exec_callback, message, consumer_backend):
        mock_exec_callback.side_effect = Exception
        with pytest.raises(mock_exec_callback.side_effect):
            consumer_backend.message_handler(*message.serialize(), None)


pre_process_hook = mock.MagicMock()
post_process_hook = mock.MagicMock()
heartbeat_hook = mock.MagicMock()


@pytest.fixture(name='prepost_process_hooks')
def _prepost_process_hooks(settings):
    settings.HEDWIG_HEARTBEAT_HOOK = 'tests.test_backends.test_base.heartbeat_hook'
    settings.HEDWIG_PRE_PROCESS_HOOK = 'tests.test_backends.test_base.pre_process_hook'
    settings.HEDWIG_POST_PROCESS_HOOK = 'tests.test_backends.test_base.post_process_hook'
    yield
    pre_process_hook.reset_mock(side_effect=True)
    post_process_hook.reset_mock(side_effect=True)
    heartbeat_hook.reset_mock(side_effect=True)


class TestFetchAndProcessMessages:
    def test_success(self, consumer_backend):
        num_messages = 3
        visibility_timeout = 4
        shutdown_event = threading.Event()

        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [mock.MagicMock(), mock.MagicMock()], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.ack_message = mock.MagicMock()

        consumer_backend.fetch_and_process_messages(num_messages, visibility_timeout, shutdown_event)

        consumer_backend.pull_messages.assert_called_with(
            num_messages=num_messages, visibility_timeout=visibility_timeout, shutdown_event=shutdown_event
        )
        consumer_backend.process_message.assert_has_calls(
            [mock.call(x) for x in consumer_backend.pull_messages.return_value]
        )
        consumer_backend.ack_message.assert_has_calls(
            [mock.call(x) for x in consumer_backend.pull_messages.return_value]
        )

    def test_preserves_messages(self, consumer_backend):
        consumer_backend.pull_messages = mock.MagicMock()
        shutdown_event = threading.Event()
        mock_return_once(consumer_backend.pull_messages, [mock.MagicMock()], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.process_message.side_effect = Exception

        consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

        consumer_backend.pull_messages.return_value[0].delete.assert_not_called()

    def test_ignore_delete_error(self, consumer_backend):
        queue_message = mock.MagicMock()
        shutdown_event = threading.Event()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.ack_message = mock.MagicMock(side_effect=Exception)

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with(
                'hedwig.backends.base', logging.ERROR, mock.ANY, exc_info=True, extra=mock.ANY
            )

        consumer_backend.ack_message.assert_called_once_with(queue_message)

    def test_pre_process_hook(self, consumer_backend, prepost_process_hooks):
        shutdown_event = threading.Event()
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [mock.MagicMock(), mock.MagicMock()], [], shutdown_event)

        consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

        pre_process_hook.assert_has_calls(
            [
                mock.call(**consumer_backend.pre_process_hook_kwargs(x))
                for x in consumer_backend.pull_messages.return_value
            ]
        )

    def test_pre_process_hook_exception(self, consumer_backend, prepost_process_hooks):
        shutdown_event = threading.Event()
        pre_process_hook.side_effect = RuntimeError('fail')
        queue_message = mock.MagicMock()
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with(
                'hedwig.backends.base',
                logging.ERROR,
                'Exception in pre process hook for message',
                exc_info=True,
                extra={'queue_message': queue_message},
            )

        pre_process_hook.assert_called_once_with(**consumer_backend.pre_process_hook_kwargs(queue_message))
        queue_message.delete.assert_not_called()

    def test_post_process_hook(self, consumer_backend, prepost_process_hooks):
        shutdown_event = threading.Event()
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [mock.MagicMock(), mock.MagicMock()], [], shutdown_event)

        consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

        post_process_hook.assert_has_calls(
            [
                mock.call(**consumer_backend.post_process_hook_kwargs(x))
                for x in consumer_backend.pull_messages.return_value
            ]
        )

    def test_post_process_hook_exception(self, consumer_backend, prepost_process_hooks):
        shutdown_event = threading.Event()
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        queue_message = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)
        post_process_hook.side_effect = RuntimeError('fail')

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with(
                'hedwig.backends.base',
                logging.ERROR,
                'Exception in post process hook for message',
                exc_info=True,
                extra={'queue_message': queue_message},
            )

        post_process_hook.assert_called_once_with(**consumer_backend.pre_process_hook_kwargs(queue_message))
        queue_message.delete.assert_not_called()

    def test_special_handling_logging_error(self, consumer_backend):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock(
            side_effect=LoggingException('foo', extra={'mickey': 'mouse'})
        )

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with(
                'hedwig.backends.base', logging.ERROR, 'foo', exc_info=True, extra={'mickey': 'mouse'}
            )

    def test_special_handling_retry_error(self, consumer_backend):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock(side_effect=RetryException)

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with('hedwig.backends.base', logging.INFO, mock.ANY)

    def test_special_handling_ignore_exception(self, consumer_backend):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)
        consumer_backend.process_message = mock.MagicMock(side_effect=IgnoreException)

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_once_with('hedwig.backends.base', logging.INFO, mock.ANY, extra=mock.ANY)

    def test_handling_exception_increase_error_count(self, consumer_backend):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(
            consumer_backend.pull_messages, [queue_message, queue_message, queue_message], [], shutdown_event
        )

        # First message is processed normally, second and third raise an exception
        consumer_backend.process_message = mock.MagicMock(side_effect=[None, Exception, Exception])

        with mock.patch('hedwig.backends.base.log') as logging_mock:
            consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

            logging_mock.assert_called_with('hedwig.backends.base', logging.ERROR, mock.ANY, exc_info=True)
            assert logging_mock.call_count == 2

        assert consumer_backend.error_count == 2

    def test_success_reset_error_count(self, consumer_backend):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(consumer_backend.pull_messages, [queue_message], [], shutdown_event)

        # start with error count at 2
        consumer_backend._error_count = 2

        # process message normally
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

        # error count is reset to 0
        assert consumer_backend.error_count == 0

    def test_heartbeat_hook_called_once_when_multiple_messages_are_processed(
        self, consumer_backend, prepost_process_hooks
    ):
        shutdown_event = threading.Event()
        queue_message = mock.MagicMock()
        consumer_backend.pull_messages = mock.MagicMock()
        mock_return_once(
            consumer_backend.pull_messages, [queue_message, queue_message, queue_message], [], shutdown_event
        )

        # process message normally
        consumer_backend.process_message = mock.MagicMock()
        consumer_backend.fetch_and_process_messages(shutdown_event=shutdown_event)

        assert pre_process_hook.call_count == 3
        assert post_process_hook.call_count == 3
        heartbeat_hook.assert_called_once_with(error_count=0)


default_headers = mock.MagicMock(return_value={'mickey': 'mouse'})


@pytest.fixture(name='default_headers_hook')
def _default_headers_hook(settings):
    settings.HEDWIG_DEFAULT_HEADERS = 'tests.test_backends.test_base.default_headers'
    yield default_headers
    default_headers.reset_mock()


def pre_serialize_hook(message_data):
    # clear headers to make sure we are not able to destroy message attributes
    message_data['metadata']['headers'].clear()


class TestPublisher:
    def test_publish(self, message, mock_publisher_backend, use_transport_message_attrs):
        mock_publisher_backend.publish(message)

        mock_publisher_backend._publish.assert_called_once_with(message, *message.serialize())

    def test_default_headers_hook(
        self, message, mock_publisher_backend, default_headers_hook, use_transport_message_attrs
    ):
        mock_publisher_backend.publish(message)

        default_headers_hook.assert_called_once_with(message=message)

        headers = {**message.headers, **default_headers_hook.return_value}
        payload, attributes = message.with_headers(headers).serialize()

        mock_publisher_backend._publish.assert_called_once_with(mock.ANY, mock.ANY, mock.ANY)
        if settings.HEDWIG_DATA_VALIDATOR_CLASS.__name__ == 'ProtobufValidator':
            from hedwig.protobuf.container_pb2 import PayloadV1  # noqa

            if not use_transport_message_attrs:
                msg = PayloadV1()
                msg.ParseFromString(mock_publisher_backend._publish.call_args[0][1])
                expected_msg = PayloadV1()
                expected_msg.ParseFromString(payload)
                assert expected_msg == msg
            else:
                assert mock_publisher_backend._publish.call_args[0][1] == payload
        else:
            assert json.loads(mock_publisher_backend._publish.call_args[0][1]) == json.loads(payload)
        if not use_transport_message_attrs:
            assert mock_publisher_backend._publish.call_args[0][2] == headers
        else:
            transport_message_attrs = mock_publisher_backend._publish.call_args[0][2]
            sent_headers = {k: v for k, v in transport_message_attrs.items() if not k.startswith("hedwig_")}
            assert sent_headers == {
                **headers,
                **default_headers_hook.return_value,
            }
            assert mock_publisher_backend._publish.call_args[0][2] == attributes



import base64
import re
import typing
from copy import deepcopy
from distutils.version import StrictVersion
from typing import Tuple, Union, TypeVar, Type, List, Optional, MutableMapping, cast

import funcy
from google.protobuf import json_format
from google.protobuf.any_pb2 import Any
from google.protobuf.message import DecodeError, Message as ProtoMessage
from google.protobuf.struct_pb2 import Value

from hedwig.conf import settings
from hedwig.exceptions import ValidationError
from hedwig.protobuf import options_pb2
from hedwig.protobuf.container_pb2 import PayloadV1
from hedwig.protobuf.options_pb2 import MessageOptions
from hedwig.validators.base import HedwigBaseValidator, MetaAttributes


class SchemaError(Exception):
    pass


ProtoMessageT = TypeVar("ProtoMessageT", bound=ProtoMessage)


def decode_proto_json(msg_class: typing.Any, value: Union[str, bytes]) -> ProtoMessageT:  # type: ignore
    assert isinstance(value, str)

    msg: ProtoMessageT = msg_class()
    json_format.Parse(value, msg, ignore_unknown_fields=True)
    return msg


def encode_proto_json(msg: ProtoMessage) -> str:
    return json_format.MessageToJson(msg, preserving_proto_field_name=True, indent=0).replace("\n", "")


class ProtobufValidator(HedwigBaseValidator):
    """
    A validator that encodes the payload using Protobuf binary format.
    """

    proto_messages: MutableMapping[Tuple[str, int], Type[ProtoMessage]]
    """
    The list of protoc compiled python classes - supplied by app
    """

    _version_pattern_re = re.compile(r"^([0-9]+)\.\*$")

    _message_name_re = re.compile(r"^(.*)V(\d+)$")

    def __init__(self, proto_messages: Optional[List[Type[ProtoMessage]]] = None) -> None:
        # schema encoding, eg: hedwig.automatic.com/schema#/schemas/trip.created/1.0
        schema_fmt = '{message_type}/{message_version}'
        schema_re = re.compile(r'([^/]+)/([^/]+)$')
        self.proto_messages = {}

        super().__init__(schema_fmt, schema_re, StrictVersion('1.0'))

        if proto_messages is None:
            proto_messages = settings.HEDWIG_PROTOBUF_MESSAGES

        self._check_schema(proto_messages)

    def _decode_proto(self, msg_class: typing.Any, value: Union[str, bytes]) -> ProtoMessageT:  # type: ignore
        """
        Decode an arbitrary protobuf message from value
        """
        assert isinstance(value, bytes)

        msg: ProtoMessageT = msg_class()
        msg.ParseFromString(value)
        return msg

    def _encode_proto(self, msg: ProtoMessage) -> Union[str, bytes]:
        """
        Encode an arbitrary protobuf message into value
        """
        return msg.SerializeToString()

    def _verify_known_minor_version(self, message_type: str, full_version: StrictVersion):
        msg_class = self.proto_messages.get((message_type, full_version.version[0]))
        if not msg_class:
            raise ValidationError(f"Protobuf message class not found for '{message_type}/{full_version.version[0]}'")

        options: MessageOptions = msg_class.DESCRIPTOR.GetOptions().Extensions[options_pb2.message_options]
        if options.minor_version < full_version.version[1]:
            raise ValidationError(
                f'Unknown minor version: {full_version.version[1]}, last known minor version: '
                f'{options.minor_version}'
            )

    def _extract_data(
        self, message_payload: Union[bytes, str], attributes: dict, use_transport_attributes: bool
    ) -> Tuple[MetaAttributes, bytes]:
        assert isinstance(message_payload, (bytes, str))

        if not use_transport_attributes:
            try:
                msg_payload = self._decode_proto(PayloadV1, message_payload)  # type: ignore
            except (DecodeError, RuntimeError, AssertionError, json_format.ParseError) as e:
                raise ValidationError(f"Invalid data for message: PayloadV1: {e}")

            data = msg_payload.data
            meta_attrs = MetaAttributes(
                msg_payload.metadata.timestamp.ToMilliseconds(),
                msg_payload.metadata.publisher,
                dict(msg_payload.metadata.headers),
                msg_payload.id,
                msg_payload.schema,
                msg_payload.format_version,
            )
        else:
            data = message_payload
            meta_attrs = self._decode_meta_attributes(attributes)
            if meta_attrs.format_version != self._current_format_version:
                raise ValidationError(f"Invalid format version: {meta_attrs.format_version}")
        return meta_attrs, data

    def _extract_data_firehose(self, line: str) -> Tuple[MetaAttributes, bytes]:
        assert isinstance(line, str)

        try:
            msg_payload = decode_proto_json(PayloadV1, line)  # type: ignore
        except (DecodeError, RuntimeError, AssertionError, json_format.ParseError) as e:
            raise ValidationError(f"Invalid data for message: PayloadV1: {e}")

        data = msg_payload.data
        if data.Is(Value.DESCRIPTOR):
            # unknown data types are encoded as `encode(PayloadV1(data=Value(string_value=base64(binary))))`
            value_msg = Value()
            data.Unpack(value_msg)
            assert value_msg.WhichOneof("kind") == "string_value"
            data = base64.decodebytes(value_msg.string_value.encode("utf8"))

        meta_attrs = MetaAttributes(
            msg_payload.metadata.timestamp.ToMilliseconds(),
            msg_payload.metadata.publisher,
            dict(msg_payload.metadata.headers),
            msg_payload.id,
            msg_payload.schema,
            msg_payload.format_version,
        )
        return meta_attrs, data

    def _decode_data(
        self,
        meta_attrs: MetaAttributes,
        message_type: str,
        full_version: StrictVersion,
        data: Union[Any, bytes, str],
    ) -> ProtoMessage:
        assert isinstance(data, (Any, bytes, str))

        msg_class = self.proto_messages.get((message_type, full_version.version[0]))
        if not msg_class:
            raise ValidationError(f"Protobuf message class not found for '{message_type}/{full_version.version[0]}'")

        try:
            if isinstance(data, Any):
                assert data.Is(msg_class.DESCRIPTOR)
                data_msg = msg_class()
                data.Unpack(data_msg)
            else:
                data_msg = self._decode_proto(msg_class, data)
        except (DecodeError, RuntimeError, AssertionError, json_format.ParseError) as e:
            raise ValidationError(f"Invalid data for message: {msg_class.__name__}: {e}")
        return data_msg

    def _encode_payload(
        self, meta_attrs: MetaAttributes, data: ProtoMessage, use_transport_attributes: bool
    ) -> Tuple[Union[bytes, str], dict]:
        assert isinstance(data, ProtoMessage)

        if not use_transport_attributes:
            msg = PayloadV1()
            msg.format_version = str(self._current_format_version)
            msg.id = str(meta_attrs.id)
            msg.metadata.publisher = meta_attrs.publisher
            msg.metadata.timestamp.FromMilliseconds(meta_attrs.timestamp)
            for k, v in meta_attrs.headers.items():
                msg.metadata.headers[k] = v
            msg.schema = meta_attrs.schema
            msg.data.Pack(data)
            payload = self._encode_proto(msg)
            msg_attrs = deepcopy(meta_attrs.headers)
        else:
            payload = self._encode_proto(data)
            msg_attrs = self._encode_meta_attributes(meta_attrs)
        return payload, msg_attrs

    def _encode_payload_firehose(
        self, message_type: str, version: StrictVersion, meta_attrs: MetaAttributes, data: ProtoMessage
    ) -> str:
        assert isinstance(data, ProtoMessage)

        msg = PayloadV1()
        msg.format_version = str(self._current_format_version)
        msg.id = str(meta_attrs.id)
        msg.metadata.publisher = meta_attrs.publisher
        msg.metadata.timestamp.FromMilliseconds(meta_attrs.timestamp)
        for k, v in meta_attrs.headers.items():
            msg.metadata.headers[k] = v
        msg.schema = meta_attrs.schema
        try:
            self._verify_known_minor_version(message_type, version)
            msg.data.Pack(data)
        except ValidationError:
            value_msg = Value()
            value_msg.string_value = base64.b64encode(data.SerializeToString()).decode()
            msg.data.Pack(value_msg)
        payload = encode_proto_json(msg)
        return payload

    def _check_schema(self, proto_messages) -> None:
        msg_types_found = {k for k in funcy.chain(settings.HEDWIG_MESSAGE_ROUTING, settings.HEDWIG_CALLBACKS)}
        errors = []
        for proto_message in proto_messages:
            message_type: Optional[str] = None
            major_version: int
            if options_pb2.message_options not in proto_message.DESCRIPTOR.GetOptions().Extensions:
                errors.append(f"Protobuf message class '{proto_message}' does not define option message_options")
            options = proto_message.DESCRIPTOR.GetOptions().Extensions[options_pb2.message_options]
            if not options.major_version:  # default is 0 which is invalid
                errors.append(
                    f"Protobuf message class '{proto_message}' does not define option message_options.major_version"
                )
                continue
            else:
                major_version = cast(int, options.major_version)
            if options.message_type:
                message_type = cast(str, options.message_type)
            m = self._message_name_re.match(proto_message.__name__)
            if m:
                name_message_type = m.group(1)
                name_major_version = int(m.group(2))
                if message_type is None:
                    message_type = name_message_type
                if name_major_version != major_version:
                    errors.append(
                        f"Protobuf message class '{proto_message}' mismatch in major version: '{name_major_version}' "
                        f"!= '{major_version}'"
                    )
            if not message_type:
                errors.append(f"Couldn't determine message type for Protobuf message class '{proto_message}'")
                continue
            if (message_type, major_version) in self.proto_messages:
                errors.append(f"Duplicate Protobuf message declared for '{message_type}/{major_version}'")
                continue
            self.proto_messages[(message_type, major_version)] = proto_message

        for message_type, version_pattern in msg_types_found:
            m = self._version_pattern_re.match(version_pattern)
            if not m:
                errors.append(f"Invalid version '{version_pattern}' for message: '{message_type}'")
                continue
            major_version = int(m.group(1))
            if major_version == 0:
                errors.append(f"Invalid version '{major_version}' for message: '{message_type}'. Must not be 0.")
            assert message_type  # mypy
            msg_class = self.proto_messages.get((message_type, major_version))
            if not msg_class:
                errors.append(
                    f"Invalid message_type, major version: '{message_type}/{major_version}', not found in declared "
                    f"messages"
                )
                continue

        if errors:
            raise SchemaError(str(errors))


class ProtobufJSONValidator(ProtobufValidator):
    """
    A validator that encodes the payload using Protobuf JSON format.
    Documentation: https://googleapis.dev/python/protobuf/latest/google/protobuf/json_format.html
    """

    def _decode_proto(self, msg_class: typing.Any, value: Union[str, bytes]) -> ProtoMessageT:  # type: ignore
        return decode_proto_json(msg_class, value)

    def _encode_proto(self, msg: ProtoMessage) -> Union[str, bytes]:
        return encode_proto_json(msg)



import threading
import typing
from typing import Optional

from hedwig.backends.utils import get_consumer_backend


def process_messages_for_lambda_consumer(lambda_event: dict) -> None:
    sns_consumer_backend = get_consumer_backend()
    sns_consumer_backend.process_messages(lambda_event)


def listen_for_messages(
    num_messages: int = 10,
    visibility_timeout_s: typing.Optional[int] = None,
    shutdown_event: Optional[threading.Event] = None,
) -> None:
    """
    Starts a Hedwig listener for message types provided and calls the callback handlers like so:

    .. code-block:: python

        callback_fn(message)

    The message is explicitly deleted only if callback function ran successfully. In case of an exception the message is
    kept on queue and processed again. If the callback keeps failing, the message is moved to the dead-letter queue.

    This function is blocking by default. It may be stopped by passing a shut down event object which can be set to
    stop the function.

    :param num_messages: Maximum number of messages to fetch in one API call. Defaults to 10
    :param visibility_timeout_s: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
    :param shutdown_event: An event to signal that the process should shut down. This prevents more messages from
        being de-queued and function exits after the current messages have been processed.
    """
    if not shutdown_event:
        shutdown_event = threading.Event()

    consumer_backend = get_consumer_backend()
    consumer_backend.fetch_and_process_messages(
        num_messages=num_messages, visibility_timeout=visibility_timeout_s, shutdown_event=shutdown_event
    )



import copy
import dataclasses
import time
import uuid
from concurrent.futures import Future
from distutils.version import StrictVersion
from enum import Enum
from functools import lru_cache
from typing import Union, Any, cast, Tuple, Dict, Optional

from hedwig.backends.utils import get_consumer_backend
from hedwig.conf import settings
from hedwig.exceptions import ValidationError, CallbackNotFound


@lru_cache(maxsize=1)
def _validator():
    return settings.HEDWIG_DATA_VALIDATOR_CLASS()


@dataclasses.dataclass(frozen=True)
class Metadata:
    timestamp: int = dataclasses.field(default_factory=lambda: int(time.time() * 1000))
    """
    Timestamp of message creation in epoch milliseconds
    """

    publisher: str = dataclasses.field(default_factory=lambda: settings.HEDWIG_PUBLISHER)
    """
    Publisher of message
    """

    headers: Dict[str, str] = dataclasses.field(default_factory=dict)
    """
    Custom headers sent with the message
    """

    provider_metadata: Any = None
    """
    Provider specific metadata, such as SQS Receipt, or Google ack id. This may be used to extend message
    visibility if the task is running longer than expected using :meth:`Message.extend_visibility_timeout`
    """


@dataclasses.dataclass(frozen=True)
class Message:
    """
    Model for Hedwig messages.
    A Message object will always have known message schema and schema version even if the data _may_ not be valid.
    """

    data: Any
    """
    Message data
    """

    type: str = dataclasses.field()
    """
    Message type. May be none if message is invalid
    """

    version: StrictVersion = dataclasses.field()
    """
    `StrictVersion` object representing data schema version.
    """

    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    """
    Message identifier
    """

    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    """
    Message metadata
    """

    @staticmethod
    def deserialize(payload: Union[str, bytes], attributes: dict, provider_metadata: Any) -> 'Message':
        """
        Deserialize a message from on-the-wire payload
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        return _validator().deserialize(payload, attributes, provider_metadata)

    @staticmethod
    def deserialize_firehose(line: str) -> 'Message':
        """
        Deserialize a message from a line of firehose file
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        return _validator().deserialize_firehose(line)

    @staticmethod
    def deserialize_containerized(payload: Union[str, bytes]) -> 'Message':
        """
        Deserialize a message assuming containerized format regardless of configured settings
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        return _validator().deserialize_containerized(payload)

    def exec_callback(self) -> None:
        """
        Call the callback with this message
        """
        self.callback.call(self)

    @classmethod
    def new(
        cls,
        msg_type: Union[str, Enum],
        version: StrictVersion,
        data: Any,
        msg_id: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> 'Message':
        """
        Creates Message object given type, data schema version and data. This is typically used by the publisher code.

        :param msg_type: message type (could be an enum, it's value will be used)
        :param version: StrictVersion representing data schema
        :param data: The dict to pass in `data` field of Message.
        :param msg_id: Custom message identifier. If not passed, a randomly generated uuid will be used.
        :param headers: Custom headers (keys must not begin with reserved namespace `hedwig_`)
        """
        assert isinstance(msg_type, (str, Enum))
        assert isinstance(version, StrictVersion)
        assert isinstance(msg_id, (type(None), str))
        assert isinstance(headers, (type(None), dict))

        if isinstance(msg_type, Enum):
            msg_type = msg_type.value

        return Message(
            id=msg_id or str(uuid.uuid4()),
            type=cast(str, msg_type),
            version=version,
            metadata=Metadata(headers=headers or {}),
            data=copy.deepcopy(data),
        )

    def publish(self) -> Union[str, Future]:
        """
        Publish this message on Hedwig infra
        :returns: for async publishers, returns a future that represents the publish api call, otherwise, returns
        the published message id
        """
        from hedwig.publisher import publish

        return publish(self)

    def extend_visibility_timeout(self, visibility_timeout_s: int) -> None:
        """
        Extends visibility timeout of a message for long running tasks.
        """
        consumer_backend = get_consumer_backend()
        consumer_backend.extend_visibility_timeout(visibility_timeout_s, self.provider_metadata)

    @property
    def callback(self):
        from hedwig.callback import Callback

        try:
            return Callback.find_by_message(self.type, self.major_version)
        except CallbackNotFound:
            raise ValidationError

    @property
    def major_version(self) -> int:
        return self.version.version[0]

    @property
    def timestamp(self) -> int:
        return self.metadata.timestamp

    @property
    def headers(self) -> Dict[str, str]:
        return self.metadata.headers

    @property
    def provider_metadata(self):
        return self.metadata.provider_metadata

    @property
    def publisher(self) -> str:
        return self.metadata.publisher

    def serialize(self) -> Tuple[Union[str, bytes], dict]:
        """
        Serialize a message for appropriate on-the-wire format
        :return: Tuple of message payload and transport attributes
        """
        return _validator().serialize(self)

    def serialize_containerized(self) -> Union[str, bytes]:
        """
        Serialize a message using containerized format regardless of configured settings. In most cases, you just want
        to use `.serialize`.
        :return: Message payload
        """
        return _validator().serialize_containerized(self)

    def serialize_firehose(self) -> str:
        """
        Serialize a message for appropriate firehose file format. See
        :meth:`hedwig.validators.base.HedwigBaseValidator.deserialize_firehose` for details.
        :return: Tuple of message payload and transport attributes
        """
        return _validator().serialize_firehose(self)

    def with_headers(self, new_headers: Dict[str, str]) -> 'Message':
        """
        Creates a copy of the message with different headers.
        :param new_headers:
        :return:
        """
        return dataclasses.replace(self, metadata=dataclasses.replace(self.metadata, headers=new_headers))

    def with_provider_metadata(self, new_provider_metadata: Any) -> 'Message':
        """
        Creates a copy of the message with different provider metadata.
        :param new_provider_metadata:
        :return:
        """
        return dataclasses.replace(
            self, metadata=dataclasses.replace(self.metadata, provider_metadata=new_provider_metadata)
        )
